from av.codec.context cimport CodecContext
from av.packet cimport Packet


cdef class SubtitleCodecContext(CodecContext):
    cdef bint subtitle_header_set
    cpdef decode(self, Packet packet=?)
    cpdef decode2(self, Packet packet)



import pytest

import av
from av.video.reformatter import (
    ColorPrimaries,
    ColorRange,
    Colorspace,
    ColorTrc,
)

from .common import fate_suite


def test_penguin_joke() -> None:
    container = av.open(
        fate_suite("amv/MTV_high_res_320x240_sample_Penguin_Joke_MTV_from_WMV.amv")
    )
    stream = container.streams.video[0]

    assert stream.codec_context.color_range == 2
    assert stream.codec_context.color_range == ColorRange.JPEG

    assert stream.codec_context.color_primaries == 2
    assert stream.codec_context.color_trc == 2

    assert stream.codec_context.colorspace == 5
    assert stream.codec_context.colorspace == Colorspace.ITU601

    for frame in container.decode(stream):
        assert frame.color_range == ColorRange.JPEG  # a.k.a "pc"
        assert frame.colorspace == Colorspace.ITU601
        return


def test_sky_timelapse() -> None:
    container = av.open(
        av.datasets.curated("pexels/time-lapse-video-of-night-sky-857195.mp4")
    )
    stream = container.streams.video[0]

    assert stream.disposition == av.stream.Disposition.default

    assert stream.codec_context.color_range == 1
    assert stream.codec_context.color_range == ColorRange.MPEG
    assert stream.codec_context.color_primaries == 1
    assert stream.codec_context.color_trc == 1
    assert stream.codec_context.colorspace == 1


def test_frame_color_trc_property() -> None:
    frame = av.VideoFrame(width=64, height=64, format="rgb24")
    assert frame.color_trc == ColorTrc.UNSPECIFIED

    frame.color_trc = ColorTrc.IEC61966_2_4
    assert frame.color_trc == ColorTrc.IEC61966_2_4

    frame.color_trc = ColorTrc.BT709
    assert frame.color_trc == ColorTrc.BT709


def test_frame_color_primaries_property() -> None:
    frame = av.VideoFrame(width=64, height=64, format="rgb24")
    assert frame.color_primaries == ColorPrimaries.UNSPECIFIED

    frame.color_primaries = ColorPrimaries.BT709
    assert frame.color_primaries == ColorPrimaries.BT709
    assert frame.color_primaries == 1  # AVCOL_PRI_BT709


def test_reformat_dst_color_trc() -> None:
    # Reformat a frame and tag it with sRGB transfer characteristic.
    frame = av.VideoFrame(width=64, height=64, format="yuv420p")
    rgb = frame.reformat(
        format="rgb24",
        dst_colorspace=Colorspace.ITU709,
        dst_color_trc=ColorTrc.IEC61966_2_4,
    )
    assert rgb.format.name == "rgb24"
    assert rgb.colorspace == Colorspace.ITU709
    assert rgb.color_trc == ColorTrc.IEC61966_2_4


def test_reformat_dst_color_primaries() -> None:
    frame = av.VideoFrame(width=64, height=64, format="yuv420p")
    rgb = frame.reformat(
        format="rgb24",
        dst_color_primaries=ColorPrimaries.BT709,
    )
    assert rgb.color_primaries == ColorPrimaries.BT709


def test_reformat_preserves_color_trc() -> None:
    # When dst_color_trc is not specified, the source frame's value is preserved.
    frame = av.VideoFrame(width=64, height=64, format="yuv420p")
    frame.color_trc = ColorTrc.BT709
    rgb = frame.reformat(format="rgb24")
    assert rgb.color_trc == ColorTrc.BT709


def test_reformat_preserves_color_primaries() -> None:
    # When dst_color_primaries is not specified, the source frame's value is preserved.
    frame = av.VideoFrame(width=64, height=64, format="yuv420p")
    frame.color_primaries = ColorPrimaries.BT709
    rgb = frame.reformat(format="rgb24")
    assert rgb.color_primaries == ColorPrimaries.BT709


@pytest.mark.parametrize(
    ("colorspace", "expected"),
    [
        (Colorspace.ITU709, Colorspace.ITU709),
        (Colorspace.FCC, Colorspace.FCC),
        (Colorspace.ITU601, 6),  # AVCOL_SPC_SMPTE170M
        (Colorspace.SMPTE240M, Colorspace.SMPTE240M),
        (Colorspace.BT2020, Colorspace.BT2020),
    ],
)
def test_reformat_dst_colorspace_metadata(
    colorspace: Colorspace, expected: Colorspace | int
) -> None:
    frame = av.VideoFrame(width=64, height=64, format="yuv420p")
    rgb = frame.reformat(format="rgb24", dst_colorspace=colorspace)
    assert rgb.colorspace == expected


# RESERVED0 (0) and RESERVED (3) primaries/transfer values, plus a couple of
# transfer functions swscale can't handle (LOG / LOG_SQRT). Real VP9 and NVDEC
# streams routinely tag frames with these. sws_scale_frame (used since 17.0.0)
# validates these fields and rejects them with EOPNOTSUPP, which regressed a
# plain reformat/to_ndarray to "rgb24" (#2208). The pre-17.0 sws_scale ignored
# them, and a transfer/primaries conversion should stay opt-in.
@pytest.mark.parametrize(
    ("color_primaries", "color_trc"),
    [
        (3, 3),  # RESERVED / RESERVED
        (0, 0),  # RESERVED0 / RESERVED0
        (3, 2),  # reserved primaries only
        (2, 3),  # reserved transfer only
        (2, 9),  # AVCOL_TRC_LOG (unsupported by swscale)
        (2, 10),  # AVCOL_TRC_LOG_SQRT (unsupported by swscale)
    ],
)
def test_reformat_unsupported_color_metadata(
    color_primaries: int, color_trc: int
) -> None:
    frame = av.VideoFrame(width=64, height=64, format="yuv420p")
    frame.colorspace = Colorspace.ITU709
    frame.color_primaries = color_primaries
    frame.color_trc = color_trc

    # Neither of these should raise OSError(EOPNOTSUPP).
    rgb = frame.reformat(format="rgb24")
    assert rgb.format.name == "rgb24"
    array = frame.to_ndarray(format="rgb24")
    assert array.shape == (64, 64, 3)

    # The reformat must not mutate the source frame's metadata.
    assert frame.color_primaries == color_primaries
    assert frame.color_trc == color_trc

    # The BT.709 matrix is still applied even though the transfer/primaries are
    # unsupported: a neutral gray must stay gray.
    gray = av.VideoFrame(width=64, height=64, format="yuv420p")
    gray.colorspace = Colorspace.ITU709
    gray.color_primaries = color_primaries
    gray.color_trc = color_trc
    for plane, value in zip(gray.planes, (128, 128, 128)):
        plane.update(bytes([value]) * plane.buffer_size)
    out = gray.to_ndarray(format="rgb24")
    assert out.min() == out.max() == out[0, 0, 0]



from av import VideoFormat

from .common import TestCase


class TestVideoFormats(TestCase):
    def test_invalid_pixel_format(self):
        with self.assertRaises(ValueError) as cm:
            VideoFormat("__unknown_pix_fmt", 640, 480)
        assert str(cm.exception) == "not a pixel format: '__unknown_pix_fmt'"

    def test_rgb24_inspection(self) -> None:
        fmt = VideoFormat("rgb24", 640, 480)
        assert fmt.name == "rgb24"
        assert len(fmt.components) == 3
        assert not fmt.is_planar
        assert not fmt.has_palette
        assert fmt.is_rgb
        assert fmt.chroma_width() == 640
        assert fmt.chroma_height() == 480
        assert fmt.chroma_width(1024) == 1024
        assert fmt.chroma_height(1024) == 1024
        for i in range(3):
            comp = fmt.components[i]
            assert comp.plane == 0
            assert comp.bits == 8
            assert not comp.is_luma
            assert not comp.is_chroma
            assert not comp.is_alpha
            assert comp.width == 640
            assert comp.height == 480

    def test_yuv420p_inspection(self) -> None:
        fmt = VideoFormat("yuv420p", 640, 480)
        assert fmt.name == "yuv420p"
        assert len(fmt.components) == 3
        self._test_yuv420(fmt)

    def _test_yuv420(self, fmt: VideoFormat) -> None:
        assert fmt.is_planar
        assert not fmt.has_palette
        assert not fmt.is_rgb
        assert fmt.chroma_width() == 320
        assert fmt.chroma_height() == 240
        assert fmt.chroma_width(1024) == 512
        assert fmt.chroma_height(1024) == 512
        for i, comp in enumerate(fmt.components):
            comp = fmt.components[i]
            assert comp.plane == i
            assert comp.bits == 8
        assert not fmt.components[0].is_chroma
        assert fmt.components[1].is_chroma
        assert fmt.components[2].is_chroma
        assert fmt.components[0].is_luma
        assert not fmt.components[1].is_luma
        assert not fmt.components[2].is_luma
        assert not fmt.components[0].is_alpha
        assert not fmt.components[1].is_alpha
        assert not fmt.components[2].is_alpha
        assert fmt.components[0].width == 640
        assert fmt.components[1].width == 320
        assert fmt.components[2].width == 320

    def test_yuva420p_inspection(self) -> None:
        fmt = VideoFormat("yuva420p", 640, 480)
        assert len(fmt.components) == 4
        self._test_yuv420(fmt)
        assert not fmt.components[3].is_chroma
        assert fmt.components[3].width == 640

    def test_gray16be_inspection(self) -> None:
        fmt = VideoFormat("gray16be", 640, 480)
        assert fmt.name == "gray16be"
        assert len(fmt.components) == 1
        assert not fmt.is_planar
        assert not fmt.has_palette
        assert not fmt.is_rgb
        assert fmt.chroma_width() == 640
        assert fmt.chroma_height() == 480
        assert fmt.chroma_width(1024) == 1024
        assert fmt.chroma_height(1024) == 1024
        comp = fmt.components[0]
        assert comp.plane == 0
        assert comp.bits == 16
        assert comp.is_luma
        assert not comp.is_chroma
        assert comp.width == 640
        assert comp.height == 480
        assert not comp.is_alpha

    def test_pal8_inspection(self) -> None:
        fmt = VideoFormat("pal8", 640, 480)
        assert len(fmt.components) == 1
        assert fmt.has_palette



import fractions
import io
import json
import struct
from typing import get_args
from unittest import SkipTest

import numpy

import av

from .common import fate_suite, sandboxed


class TestDataStreams:
    def generate_container_with_data_packets(self):
        file = io.BytesIO()
        packet_datas_expected = dict[fractions.Fraction, bytes]()

        container = av.open(file, format="mp4", mode="w")
        stream = container.add_data_stream("bin_data")
        vstream = container.add_stream("h264")

        for i in range(10):
            packet_data = json.dumps("This may not work", ensure_ascii=False).encode(
                "utf-8"
            )
            if i % 2 == 0:
                packet_data = b"This works fine"
            packet = av.Packet(packet_data)
            vframe = av.VideoFrame.from_ndarray(
                numpy.random.randint(0, 255, size=(120, 120, 3), dtype=numpy.uint8)
            )
            packet.pts = i
            packet.dts = i
            packet.duration = 1
            packet.stream = stream
            container.mux(packet)
            container.mux(vstream.encode(vframe))
            # NOTE test passes if this is used
            # packet_datas_expected[packet.pts * packet.time_base] = bytes(packet_data)

        container.close()

        container = av.open(file, format="mp4", mode="r")
        return container, packet_datas_expected

    def test_data_packet_bytes(self):
        container, packet_datas_expected = self.generate_container_with_data_packets()

        packet_datas = dict[fractions.Fraction, bytes]()
        for packet in container.demux(container.streams.data[0]):
            if packet.pts is None:
                continue
            assert bytes(packet) in (
                b"This works fine",
                json.dumps("This may not work", ensure_ascii=False).encode("utf-8"),
            ), bytes(packet)
            # NOTE test passes if this is used
            # packet_datas[packet.pts * packet.time_base] = bytes(packet)

        assert packet_datas == packet_datas_expected


class TestProperties:
    def test_is_keyframe(self) -> None:
        with av.open(fate_suite("h264/interlaced_crop.mp4")) as container:
            stream = container.streams.video[0]
            for i, packet in enumerate(container.demux(stream)):
                if i in (0, 21, 45, 69, 93, 117):
                    assert packet.is_keyframe
                else:
                    assert not packet.is_keyframe

    def test_is_corrupt(self) -> None:
        with av.open(fate_suite("mov/white_zombie_scrunch-part.mov")) as container:
            stream = container.streams.video[0]
            for i, packet in enumerate(container.demux(stream)):
                if i == 65:
                    assert packet.is_corrupt
                else:
                    assert not packet.is_corrupt

    def test_is_discard(self) -> None:
        with av.open(fate_suite("mov/mov-1elist-ends-last-bframe.mov")) as container:
            stream = container.streams.video[0]
            for i, packet in enumerate(container.demux(stream)):
                if i == 46:
                    assert packet.is_discard
                else:
                    assert not packet.is_discard

    def test_is_disposable(self) -> None:
        with av.open(fate_suite("hap/HAPQA_NoSnappy_127x1.mov")) as container:
            stream = container.streams.video[0]
            for i, packet in enumerate(container.demux(stream)):
                if i == 0:
                    assert packet.is_disposable
                else:
                    assert not packet.is_disposable

    def test_set_duration(self) -> None:
        with av.open(fate_suite("h264/interlaced_crop.mp4")) as container:
            for packet in container.demux():
                assert packet.duration is not None
                old_duration = packet.duration
                packet.duration += 10

                assert packet.duration == old_duration + 10


class TestPacketSideData:
    def test_data_types(self) -> None:
        dtypes = get_args(av.packet.PktSideDataT)

        if av.ffmpeg_version_info.startswith("n") or av.ffmpeg_version_info.count(
            "."
        ) not in (1, 2):
            raise SkipTest(f"Expect version to be SemVar: {av.ffmpeg_version_info}")

        ffmpeg_ver = [int(v) for v in av.ffmpeg_version_info.split(".", 2)[:2]]
        for dtype in dtypes:
            av_enum = av.packet.packet_sidedata_type_from_literal(dtype)
            assert dtype == av.packet.packet_sidedata_type_to_literal(av_enum)

            if (ffmpeg_ver[0] < 8 and dtype == "lcevc") or (
                ffmpeg_ver[0] < 9 and dtype == "rtcp_sr"
            ):
                break

    def test_iter(self) -> None:
        with av.open(fate_suite("h264/extradata-reload-multi-stsd.mov")) as container:
            for pkt in container.demux():
                for sdata in pkt.iter_sidedata():
                    assert pkt.dts == 2 and sdata.data_type == "new_extradata"

    def test_palette(self) -> None:
        with av.open(fate_suite("h264/extradata-reload-multi-stsd.mov")) as container:
            iterpackets = container.demux()
            pkt = next(pkt for pkt in iterpackets if pkt.has_sidedata("new_extradata"))

            sdata = pkt.get_sidedata("new_extradata")
            assert sdata.data_type == "new_extradata"
            assert bool(sdata)
            assert sdata.data_size > 0
            assert sdata.data_desc == "New Extradata"

            nxt = next(iterpackets)  # has no palette

            assert not nxt.has_sidedata("new_extradata")

            sdata1 = nxt.get_sidedata("new_extradata")
            assert sdata1.data_type == "new_extradata"
            assert not bool(sdata1)
            assert sdata1.data_size == 0

            nxt.set_sidedata(sdata, move=True)
            assert not bool(sdata)

    def test_buffer_protocol(self) -> None:
        with av.open(fate_suite("h264/extradata-reload-multi-stsd.mov")) as container:
            for pkt in container.demux():
                if pkt.has_sidedata("new_extradata"):
                    sdata = pkt.get_sidedata("new_extradata")

                    raw = bytes(sdata)
                    assert len(raw) == sdata.data_size > 0
                    assert sdata.buffer_size == sdata.data_size
                    assert sdata.buffer_ptr != 0
                    assert bytes(memoryview(sdata)) == raw

                    # Modify and verify changes stick
                    sdata.update(b"\xde\xad\xbe\xef" + raw[4:])
                    assert bytes(sdata)[:4] == b"\xde\xad\xbe\xef"

                    pkt.set_sidedata(sdata)
                    assert (
                        bytes(pkt.get_sidedata("new_extradata"))[:4]
                        == b"\xde\xad\xbe\xef"
                    )
                    return

        raise AssertionError("No packet with new_extradata side data found")

    def test_skip_samples_remux(self) -> None:
        # Source file has skip_end=706 on last packet. Setting to 0 should
        # result in 706 more decoded samples. And the file duration reported by
        # the container should also increase.
        output_path = sandboxed("skip_samples_modified.mkv")

        with av.open(fate_suite("mkv/codec_delay_opus.mkv")) as c:
            original_samples = sum(f.samples for f in c.decode(c.streams.audio[0]))

        with av.open(fate_suite("mkv/codec_delay_opus.mkv")) as inp:
            original_duration = inp.duration
            audio_stream = inp.streams.audio[0]
            with av.open(output_path, "w") as out:
                out_stream = out.add_stream_from_template(audio_stream)
                for pkt in inp.demux(audio_stream):
                    if pkt.size == 0:
                        continue
                    if pkt.has_sidedata("skip_samples"):
                        sdata = pkt.get_sidedata("skip_samples")
                        raw = bytes(sdata)
                        skip_end = struct.unpack("<I", raw[4:8])[0]
                        if skip_end > 0:
                            assert skip_end == 706
                            sdata.update(raw[:4] + struct.pack("<I", 0) + raw[8:])
                            pkt.set_sidedata(sdata)
                    pkt.stream = out_stream
                    out.mux(pkt)

        with av.open(output_path) as c:
            modified_samples = sum(f.samples for f in c.decode(c.streams.audio[0]))
            modified_duration = c.duration

        assert modified_samples - original_samples == 706

        assert original_duration is not None
        assert modified_duration is not None
        assert modified_duration > original_duration



from libc.stdint cimport int64_t, uint64_t


cdef extern from "libavformat/avformat.h" nogil:
    cdef int avformat_version()
    cdef char* avformat_configuration()
    cdef char* avformat_license()

    cdef int AV_TIME_BASE
    cdef int AVSEEK_FLAG_BACKWARD
    cdef int AVSEEK_FLAG_BYTE
    cdef int AVSEEK_FLAG_ANY
    cdef int AVSEEK_FLAG_FRAME
    cdef int AVIO_FLAG_WRITE

    cdef enum AVMediaType:
        AVMEDIA_TYPE_UNKNOWN
        AVMEDIA_TYPE_VIDEO
        AVMEDIA_TYPE_AUDIO
        AVMEDIA_TYPE_DATA
        AVMEDIA_TYPE_SUBTITLE
        AVMEDIA_TYPE_ATTACHMENT

    cdef struct AVStream:
        int index
        int id
        int disposition
        AVDiscard discard
        AVCodecParameters *codecpar
        AVRational time_base
        int64_t start_time
        int64_t duration
        int64_t nb_frames
        int64_t cur_dts
        AVDictionary *metadata
        AVRational avg_frame_rate
        AVRational r_frame_rate
        AVRational sample_aspect_ratio

    cdef struct AVChapter:
        int64_t id
        int64_t start
        int64_t end
        AVRational time_base
        AVDictionary *metadata

    # http://ffmpeg.org/doxygen/trunk/structAVIOContext.html
    cdef struct AVIOContext:
        unsigned char* buffer
        int buffer_size
        int write_flag
        int direct
        int seekable
        int max_packet_size
        void *opaque

    # http://ffmpeg.org/doxygen/trunk/structAVIOInterruptCB.html
    cdef struct AVIOInterruptCB:
        int (*callback)(void*)
        void *opaque

    cdef int AVIO_SEEKABLE_NORMAL
    cdef int AVSEEK_SIZE

    cdef AVIOContext* avio_alloc_context(
        unsigned char *buffer,
        int buffer_size,
        int write_flag,
        void *opaque,
        int(*read_packet)(void *opaque, uint8_t *buf, int buf_size),
        int(*write_packet)(void *opaque, const uint8_t *buf, int buf_size),
        int64_t(*seek)(void *opaque, int64_t offset, int whence)
    )

    # https://ffmpeg.org/doxygen/trunk/structAVInputFormat.html
    cdef struct AVInputFormat:
        const char *name
        const char *long_name
        const char *extensions
        int flags
        const AVClass *priv_class

    # https://ffmpeg.org/doxygen/trunk/structAVOutputFormat.html
    cdef struct AVOutputFormat:
        const char *name
        const char *long_name
        const char *extensions
        AVCodecID audio_codec
        AVCodecID video_codec
        AVCodecID subtitle_codec
        int flags
        const AVClass *priv_class

    int avformat_query_codec(const AVOutputFormat *oformat, AVCodecID codec_id, int std_compliance)

    # AVInputFormat.flags and AVOutputFormat.flags
    cdef enum:
        AVFMT_NOFILE
        AVFMT_NEEDNUMBER
        AVFMT_SHOW_IDS
        AVFMT_GLOBALHEADER
        AVFMT_NOTIMESTAMPS
        AVFMT_GENERIC_INDEX
        AVFMT_TS_DISCONT
        AVFMT_VARIABLE_FPS
        AVFMT_NODIMENSIONS
        AVFMT_NOSTREAMS
        AVFMT_NOBINSEARCH
        AVFMT_NOGENSEARCH
        AVFMT_NO_BYTE_SEEK
        AVFMT_TS_NONSTRICT
        AVFMT_TS_NEGATIVE
        AVFMT_SEEK_TO_PTS

    # AVFormatContext.flags
    cdef enum:
        AVFMT_FLAG_GENPTS
        AVFMT_FLAG_IGNIDX
        AVFMT_FLAG_NONBLOCK
        AVFMT_FLAG_IGNDTS
        AVFMT_FLAG_NOFILLIN
        AVFMT_FLAG_NOPARSE
        AVFMT_FLAG_NOBUFFER
        AVFMT_FLAG_CUSTOM_IO
        AVFMT_FLAG_DISCARD_CORRUPT
        AVFMT_FLAG_FLUSH_PACKETS
        AVFMT_FLAG_BITEXACT
        AVFMT_FLAG_SORT_DTS
        AVFMT_FLAG_FAST_SEEK
        AVFMT_FLAG_AUTO_BSF

    cdef int av_find_best_stream(
        AVFormatContext *ic,
        AVMediaType type,
        int wanted_stream_nb,
        int related_stream,
        AVCodec **decoder_ret,
        int flags
    ) 

    cdef const AVInputFormat* av_find_input_format(const char *name)

    # http://ffmpeg.org/doxygen/trunk/structAVFormatContext.html
    cdef struct AVFormatContext:
        unsigned int nb_streams
        AVStream **streams
        unsigned int nb_chapters
        AVChapter **chapters
        const AVInputFormat *iformat
        const AVOutputFormat *oformat
        AVIOContext *pb
        AVIOInterruptCB interrupt_callback
        AVDictionary *metadata
        char filename
        int64_t start_time
        int64_t duration
        int64_t bit_rate
        int flags
        AVCodecID audio_codec_id
        AVCodecID video_codec_id
        void *opaque
        int (*io_open)(
            AVFormatContext *s,
            AVIOContext **pb,
            const char *url,
            int flags,
            AVDictionary **options
        )
        int (*io_close2)(AVFormatContext *s, AVIOContext *pb)

    cdef AVFormatContext* avformat_alloc_context()
    cdef int avformat_open_input(
        AVFormatContext **ctx,
        char *filename,
        const AVInputFormat *format,
        AVDictionary **options
    )

    cdef int avformat_close_input(AVFormatContext **ctx)
    cdef int avformat_write_header(AVFormatContext *ctx, AVDictionary **options)
    cdef int av_write_trailer(AVFormatContext *ctx)
    cdef int av_interleaved_write_frame(AVFormatContext *ctx, AVPacket *pkt)
    cdef int av_write_frame(AVFormatContext *ctx, AVPacket *pkt)
    cdef int avio_open(AVIOContext **s, char *url, int flags)
    cdef int64_t avio_size(AVIOContext *s)
    cdef const AVOutputFormat* av_guess_format(
        char *short_name, char *filename, char *mime_type
    )
    cdef int avformat_query_codec(
        AVOutputFormat *ofmt, AVCodecID codec_id, int std_compliance
    )
    cdef void avio_flush(AVIOContext *s)
    cdef int avio_close(AVIOContext *s)
    cdef int avio_closep(AVIOContext **s)
    cdef int avformat_find_stream_info(AVFormatContext *ctx, AVDictionary **options)
    cdef AVStream* avformat_new_stream(AVFormatContext *ctx, const AVCodec *c)
    cdef int avformat_alloc_output_context2(
        AVFormatContext **ctx,
        const AVOutputFormat *oformat,
        char *format_name,
        char *filename
    )
    cdef int avformat_free_context(AVFormatContext *ctx)
    cdef AVClass* avformat_get_class()
    cdef void av_dump_format(AVFormatContext *ctx, int index, char *url, int is_output)
    cdef int av_read_frame(AVFormatContext *ctx, AVPacket *packet)
    cdef int av_seek_frame(
        AVFormatContext *ctx, int stream_index, int64_t timestamp, int flags
    )
    cdef int avformat_seek_file(
        AVFormatContext *ctx,
        int stream_index,
        int64_t min_ts,
        int64_t ts,
        int64_t max_ts,
        int flags
    )
    cdef AVRational av_guess_frame_rate(
        AVFormatContext *ctx, AVStream *stream, AVFrame *frame
    )
    cdef AVRational av_guess_sample_aspect_ratio(
        AVFormatContext *ctx, AVStream *stream, AVFrame *frame
    )
    cdef const AVInputFormat* av_demuxer_iterate(void **opaque)
    cdef const AVOutputFormat* av_muxer_iterate(void **opaque)

    cdef set pyav_get_available_formats()

    cdef struct AVIndexEntry:
        int64_t pos
        int64_t timestamp
        int flags
        int size
        int min_distance

    cdef enum:
        AVINDEX_KEYFRAME
        AVINDEX_DISCARD_FRAME

    cdef AVIndexEntry *avformat_index_get_entry(AVStream *st, int idx)
    cdef int avformat_index_get_entries_count(AVStream *st)
    cdef int av_index_search_timestamp(AVStream *st, int64_t timestamp, int flags)



import os
import sys

from docutils import nodes
from sphinx.util.docutils import SphinxDirective

sys.path.insert(0, os.path.abspath(".."))


# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.extlinks",
    "sphinx.ext.doctest",
    "sphinx_copybutton",  # Add copy button to code blocks
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "PyAV"
copyright = "2026, The PyAV Team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

about = {}
with open("../av/about.py") as fp:
    exec(fp.read(), about)

release = about["__version__"]
version = release.split("-")[0]
exclude_patterns = ["_build"]
pygments_style = "sphinx"

# -- Options for HTML output ---------------------------------------------------

html_theme = "pyav"
html_theme_path = [os.path.abspath(os.path.join(__file__, "..", "_themes"))]

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logo.webp"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_theme_options = {
    "sidebarwidth": "250px",
}


doctest_global_setup = """
import errno
import os

import av
from av.datasets import fate, fate as fate_suite, curated

from tests import common
from tests.common import sandboxed as _sandboxed

def sandboxed(*args, **kwargs):
    kwargs['timed'] = True
    return _sandboxed('docs', *args, **kwargs)

_cwd = os.getcwd()
here = sandboxed('__cwd__')
try:
    os.makedirs(here)
except OSError as e:
    if e.errno != errno.EEXIST:
        raise
os.chdir(here)

video_path = curated('pexels/time-lapse-video-of-night-sky-857195.mp4')
"""

doctest_global_cleanup = "os.chdir(_cwd)"
doctest_test_doctest_blocks = ""

extlinks = {
    "issue": ("https://github.com/PyAV-Org/PyAV/issues/%s", "#%s"),
    "pr": ("https://github.com/PyAV-Org/PyAV/pull/%s", "#%s"),
    "gh-user": ("https://github.com/%s", "@%s"),
}

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

autodoc_member_order = "bysource"
autodoc_default_options = {
    "undoc-members": True,
    "show-inheritance": True,
}


todo_include_todos = True


class PyInclude(SphinxDirective):
    has_content = True

    def run(self):
        source = "\n".join(self.content)
        output = []

        def write(*content, sep=" ", end="\n"):
            output.append(sep.join(map(str, content)) + end)

        namespace = dict(write=write)
        exec(compile(source, "<docs>", "exec"), namespace, namespace)

        output = "".join(output).splitlines()
        self.state_machine.insert_input(output, "blah")

        return []  # [nodes.literal('hello', repr(content))]


def load_entrypoint(name):
    parts = name.split(":")
    if len(parts) == 1:
        parts = name.rsplit(".", 1)
    mod_name, attrs = parts

    attrs = attrs.split(".")
    try:
        obj = __import__(mod_name, fromlist=["."])
    except ImportError as e:
        print("Error while importing.", (name, mod_name, attrs, e))
        raise

    for attr in attrs:
        obj = getattr(obj, attr)

    return obj


class EnumTable(SphinxDirective):
    required_arguments = 1
    option_spec = {
        "class": lambda x: x,
    }

    def run(self):
        cls_ep = self.options.get("class")
        cls = load_entrypoint(cls_ep) if cls_ep else None

        enum = load_entrypoint(self.arguments[0])
        properties = {}

        if cls is not None:
            for name, value in vars(cls).items():
                if isinstance(value, property):
                    try:
                        item = value._enum_item
                    except AttributeError:
                        pass
                    else:
                        if isinstance(item, enum):
                            properties[item] = name

        colwidths = [15, 15, 5, 65] if cls else [15, 5, 75]
        ncols = len(colwidths)

        table = nodes.table()

        tgroup = nodes.tgroup(cols=ncols)
        table += tgroup

        for width in colwidths:
            tgroup += nodes.colspec(colwidth=width)

        thead = nodes.thead()
        tgroup += thead

        tbody = nodes.tbody()
        tgroup += tbody

        def makerow(*texts):
            row = nodes.row()
            for text in texts:
                if text is None:
                    continue
                row += nodes.entry("", nodes.paragraph("", str(text)))
            return row

        thead += makerow(
            f"{cls.__name__} Attribute" if cls else None,
            f"{enum.__name__} Name",
            "Flag Value",
            "Meaning in FFmpeg",
        )

        seen = set()
        enum_items = [
            (name, item) for name, item in vars(enum).items() if isinstance(item, enum)
        ]
        for name, item in enum_items:
            if name.lower() in seen:
                continue
            seen.add(name.lower())

            try:
                attr = properties[item]
            except KeyError:
                if cls:
                    continue
                attr = None

            value = f"0x{item.value:X}"
            doc = enum.__annotations__.get(name, "---")[1:-1]
            tbody += makerow(attr, name, value, doc)

        return [table]


def ffmpeg_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Custom role for FFmpeg API links.
    Converts :ffmpeg:`AVSomething` into proper FFmpeg API documentation links.
    """

    base_url = "https://ffmpeg.org/doxygen/8.0/struct{}.html"

    try:
        struct_name, member = text.split(".")
    except Exception:
        struct_name = None

    if struct_name is None:
        fragment = {
            "avformat_seek_file": "group__lavf__decoding.html#ga3b40fc8d2fda6992ae6ea2567d71ba30",
            "av_find_best_stream": "avformat_8c.html#a8d4609a8f685ad894c1503ffd1b610b4",
            "av_frame_make_writable": "group__lavu__frame.html#gadd5417c06f5a6b419b0dbd8f0ff363fd",
            "avformat_write_header": "group__lavf__encoding.html#ga18b7b10bb5b94c4842de18166bc677cb",
            "av_guess_frame_rate": "group__lavf__misc.html#ga698e6aa73caa9616851092e2be15875d",
            "av_guess_sample_aspect_ratio": "group__lavf__misc.html#gafa6fbfe5c1bf6792fd6e33475b6056bd",
        }.get(text, f"struct{text}.html")
        url = "https://ffmpeg.org/doxygen/8.0/" + fragment
    else:
        fragment = {
            "AVCodecContext.thread_count": "#aa852b6227d0778b62e9cc4034ad3720c",
            "AVCodecContext.thread_type": "#a7651614f4309122981d70e06a4b42fcb",
            "AVCodecContext.skip_frame": "#af869b808363998c80adf7df6a944a5a6",
            "AVCodecContext.qmin": "#a3f63bc9141e25bf7f1cda0cef7cd4a60",
            "AVCodecContext.qmax": "#ab015db3b7fcd227193a7c17283914187",
            "AVCodec.capabilities": "#af51f7ff3dac8b730f46b9713e49a2518",
            "AVCodecDescriptor.props": "#a9949288403a12812cd6e3892ac45f40f",
            "AVCodecContext.bits_per_coded_sample": "#a3866500f51fabfa90faeae894c6e955c",
            "AVFrame.color_range": "#a853afbad220bbc58549b4860732a3aa5",
            "AVFrame.color_primaries": "#a59a3f830494f2ed1133103a1bc9481e7",
            "AVFrame.color_trc": "#ab09abb126e3922bc1d010cf044087939",
            "AVFrame.colorspace": "#a9262c231f1f64869439b4fe587fe1710",
        }.get(text, f"#{member}")

        url = base_url.format(struct_name) + fragment

    node = nodes.reference(rawtext, text, refuri=url, **options)
    return [node], []


def setup(app):
    app.add_css_file("custom.css")
    app.add_role("ffmpeg", ffmpeg_role)
    app.add_directive("flagtable", EnumTable)
    app.add_directive("enumtable", EnumTable)
    app.add_directive("pyinclude", PyInclude)

    return {
        "version": "1.0",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import av

av.logging.set_level(av.logging.VERBOSE)

input_ = av.open("resources/webvtt.mkv")
output = av.open("remuxed.vtt", "w")

in_stream = input_.streams.subtitles[0]
out_stream = output.add_stream_from_template(in_stream)

for packet in input_.demux(in_stream):
    if packet.size == 0:
        continue
    packet.stream = out_stream
    output.mux(packet)

input_.close()
output.close()

print("Remuxing done")

with av.open("remuxed.vtt") as f:
    for subset in f.decode(subtitles=0):
        for sub in subset:
            print(sub.ass)



import av
import av.datasets

content = av.datasets.curated("pexels/time-lapse-video-of-night-sky-857195.mp4")
with av.open(content) as container:
    # Signal that we only want to look at keyframes.
    stream = container.streams.video[0]
    stream.codec_context.skip_frame = "NONKEY"

    for i, frame in enumerate(container.decode(stream)):
        print(frame)
        frame.to_image().save(f"night-sky.{i:04d}.jpg", quality=80)



import io
import os
import struct

import pytest

import av
import av.datasets

from .common import fate_suite


def _crc32_mpeg(data: bytes) -> int:
    crc = 0xFFFFFFFF
    for byte in data:
        crc ^= byte << 24
        for _ in range(8):
            if crc & 0x80000000:
                crc = (crc << 1) ^ 0x04C11DB7
            else:
                crc <<= 1
            crc &= 0xFFFFFFFF
    return crc


def _make_ts_packet(pid: int, payload: bytes, pusi: bool = False, cc: int = 0) -> bytes:
    pid_hi = (pid >> 8) & 0x1F
    pid_lo = pid & 0xFF
    if pusi:
        pid_hi |= 0x40
    header = bytes([0x47, pid_hi, pid_lo, 0x10 | (cc & 0x0F)])
    pkt = header + payload
    return (pkt + b"\xff" * (188 - len(pkt)))[:188]


def _make_unknown_stream_ts() -> bytes:
    """Build a minimal MPEG-TS byte string whose only stream has AVMEDIA_TYPE_UNKNOWN.

    Uses stream_type=0x82 in the PMT, which FFmpeg does not map to any known
    media type, resulting in ``stream.type == "unknown"``.
    """
    pat_data = bytes(
        [0x00, 0xB0, 0x0D, 0x00, 0x01, 0xC1, 0x00, 0x00, 0x00, 0x01, 0xE1, 0x00]
    )
    pat_data += struct.pack(">I", _crc32_mpeg(pat_data))
    pat_payload = bytes([0x00]) + pat_data

    # fmt: off
    pmt_data = bytes([0x02, 0xB0, 0x12, 0x00, 0x01, 0xC1, 0x00, 0x00,
                      0xE1, 0x02, 0xF0, 0x00, 0x82, 0xE1, 0x02, 0xF0, 0x00,
    ])
    # fmt: on
    pmt_data += struct.pack(">I", _crc32_mpeg(pmt_data))
    pmt_payload = bytes([0x00]) + pmt_data

    pes_header = bytes([0x00, 0x00, 0x01, 0xBD, 0x00, 0x0A, 0x80, 0x00, 0x00])
    pes_data = pes_header + b"\xaa" * (184 - len(pes_header))

    packets: list[bytes] = []
    for i in range(2):
        packets.append(_make_ts_packet(0x0000, pat_payload, pusi=True, cc=i))
    for i in range(2):
        packets.append(_make_ts_packet(0x0100, pmt_payload, pusi=True, cc=i))
    for i in range(200):
        packets.append(_make_ts_packet(0x0102, pes_data, pusi=(i % 10 == 0), cc=i % 16))

    return b"".join(packets)


class TestStreams:
    @pytest.fixture(autouse=True)
    def cleanup(self):
        yield
        for file in (
            "data.ts",
            "data_source.ts",
            "data_copy.ts",
            "data_with_codec.ts",
            "data_invalid.ts",
            "out.mkv",
            "video_with_attachment.mkv",
            "remuxed_attachment.mkv",
        ):
            if os.path.exists(file):
                os.remove(file)

    def test_stream_tuples(self) -> None:
        for fate_name in ("h264/interlaced_crop.mp4",):
            container = av.open(fate_suite(fate_name))

            video_streams = tuple([s for s in container.streams if s.type == "video"])
            assert video_streams == container.streams.video

            audio_streams = tuple([s for s in container.streams if s.type == "audio"])
            assert audio_streams == container.streams.audio

    def test_loudnorm(self) -> None:
        container = av.open(
            fate_suite("amv/MTV_high_res_320x240_sample_Penguin_Joke_MTV_from_WMV.amv")
        )
        audio = container.streams.audio[0]
        stats = av.filter.loudnorm.stats("i=-24.0:lra=7.0:tp=-2.0", audio)

        assert isinstance(stats, bytes) and len(stats) > 30
        assert b"inf" not in stats
        assert b'"input_i"' in stats

    def test_selection(self) -> None:
        container = av.open(
            fate_suite("amv/MTV_high_res_320x240_sample_Penguin_Joke_MTV_from_WMV.amv")
        )
        video = container.streams.video[0]

        video.thread_type = av.codec.context.ThreadType.AUTO
        assert video.thread_type == av.codec.context.ThreadType.AUTO

        video.thread_type = 0x03
        assert video.thread_type == av.codec.context.ThreadType.AUTO

        video.thread_type = "AUTO"
        assert video.thread_type == av.codec.context.ThreadType.AUTO

        audio = container.streams.audio[0]

        assert [video] == container.streams.get(video=0)
        assert [video] == container.streams.get(video=(0,))

        assert video == container.streams.best("video")
        assert audio == container.streams.best("audio")

        container = av.open(fate_suite("sub/MovText_capability_tester.mp4"))
        subtitle = container.streams.subtitles[0]
        assert subtitle == container.streams.best("subtitle")

        container = av.open(fate_suite("mxf/track_01_v02.mxf"))
        data = container.streams.data[0]
        assert data == container.streams.best("data")

    def test_discard(self) -> None:
        from av.stream import Discard

        container = av.open(
            fate_suite("amv/MTV_high_res_320x240_sample_Penguin_Joke_MTV_from_WMV.amv")
        )
        audio = container.streams.audio[0]

        # Default discard policy.
        assert audio.discard == Discard.default

        # Setter accepts the enum and round-trips.
        audio.discard = Discard.all
        assert audio.discard == Discard.all

        audio.discard = Discard.nonkey
        assert audio.discard == Discard.nonkey
        container.close()

        # Discarding a stream makes demux skip (almost) all of its packets.
        def audio_packets(discard: Discard | None) -> int:
            c = av.open(
                fate_suite(
                    "amv/MTV_high_res_320x240_sample_Penguin_Joke_MTV_from_WMV.amv"
                )
            )
            if discard is not None:
                c.streams.audio[0].discard = discard
            count = sum(
                1 for p in c.demux() if p.dts is not None and p.stream.type == "audio"
            )
            c.close()
            return count

        baseline = audio_packets(None)
        discarded = audio_packets(Discard.all)
        assert baseline > 0
        assert discarded < baseline

    def test_printing_video_stream(self) -> None:
        input_ = av.open(
            fate_suite("amv/MTV_high_res_320x240_sample_Penguin_Joke_MTV_from_WMV.amv")
        )
        container = av.open("out.mkv", "w")

        video_stream = container.add_stream("h264", rate=30)
        encoder = video_stream.codec.name

        video_stream.width = input_.streams.video[0].width
        video_stream.height = input_.streams.video[0].height
        video_stream.pix_fmt = "yuv420p"

        for frame in input_.decode(video=0):
            container.mux(video_stream.encode(frame))
            break

        repr = f"{video_stream}"
        assert repr.startswith(f"<av.VideoStream #0 {encoder}, yuv420p 160x120 at ")
        assert repr.endswith(">")

        container.close()
        input_.close()

    def test_printing_video_stream2(self) -> None:
        input_ = av.open(fate_suite("h264/interlaced_crop.mp4"))
        input_stream = input_.streams.video[0]
        container = av.open("out.mkv", "w")

        video_stream = container.add_stream_from_template(input_stream)
        encoder = video_stream.codec.name

        for frame in input_.decode(video=0):
            container.mux(video_stream.encode(frame))
            break

        repr = f"{video_stream}"
        assert repr.startswith(f"<av.VideoStream #0 {encoder}, yuv420p 640x360 at ")
        assert repr.endswith(">")

        container.close()
        input_.close()

    def test_data_stream(self) -> None:
        # First test writing and reading a simple data stream
        container1 = av.open("data.ts", "w")
        data_stream = container1.add_data_stream()

        test_data = [b"test data 1", b"test data 2", b"test data 3"]
        for i, data_ in enumerate(test_data):
            packet = av.Packet(data_)
            packet.pts = i
            packet.stream = data_stream
            container1.mux(packet)
        container1.close()

        # Test reading back the data stream
        container = av.open("data.ts")

        # Test best stream selection
        data = container.streams.best("data")
        assert data == container.streams.data[0]

        # Test get method
        assert [data] == container.streams.get(data=0)
        assert [data] == container.streams.get(data=(0,))

        # Verify we can read back all the packets, ignoring empty ones
        packets = [p for p in container.demux(data) if bytes(p)]
        assert len(packets) == len(test_data)
        for read_packet, original_data in zip(packets, test_data):
            assert bytes(read_packet) == original_data

        # Test string representation
        repr = f"{data_stream}"
        assert repr.startswith("<av.DataStream #0") and repr.endswith(">")

        container.close()

    def test_data_stream_from_template(self) -> None:
        source_path = "data_source.ts"
        payloads = [b"payload-a", b"payload-b", b"payload-c"]

        with av.open(source_path, "w") as source:
            source_stream = source.add_data_stream()
            for i, payload in enumerate(payloads):
                packet = av.Packet(payload)
                packet.pts = i
                packet.stream = source_stream
                source.mux(packet)

        copied_payloads: list[bytes] = []

        with av.open(source_path) as input_container:
            input_data_stream = input_container.streams.data[0]

            with av.open("data_copy.ts", "w") as output_container:
                output_data_stream = output_container.add_stream_from_template(
                    input_data_stream
                )

                for data_packet in input_container.demux(input_data_stream):
                    payload = bytes(data_packet)
                    if not payload:
                        continue
                    copied_payloads.append(payload)
                    clone = av.Packet(payload)
                    clone.pts = data_packet.pts
                    clone.dts = data_packet.dts
                    clone.time_base = data_packet.time_base
                    clone.stream = output_data_stream
                    output_container.mux(clone)

        with av.open("data_copy.ts") as remuxed:
            output_stream = remuxed.streams.data[0]
            assert output_stream.codec_context is None

            remuxed_payloads: list[bytes] = []
            for data_packet in remuxed.demux(output_stream):
                payload = bytes(data_packet)
                if payload:
                    remuxed_payloads.append(payload)

        assert remuxed_payloads == copied_payloads

    def test_data_stream_with_codec(self) -> None:
        """Test adding a data stream with a specific codec name."""
        # Test that invalid codec names raise appropriate errors
        with pytest.raises(ValueError, match="Unknown data codec"):
            container = av.open("data_invalid.ts", "w")
            try:
                container.add_data_stream("not_a_real_codec_name_12345")
            finally:
                container.close()

        # Test that add_data_stream with codec parameter works correctly
        # We use "bin_data" which is a data codec that's always available
        output_path = "data_with_codec.ts"
        with av.open(output_path, "w") as container:
            # Try to create a data stream with a codec
            # bin_data is a simple passthrough codec for binary data
            data_stream = container.add_data_stream("bin_data")
            klv_stream = container.add_data_stream("klv")

            assert data_stream.type == "data"
            assert klv_stream.type == "data"
            # Note: codec_context may be None for descriptor-only data codecs

            test_data = [b"test1", b"test2", b"test3"]
            for i, data in enumerate(test_data):
                packet = av.Packet(data)
                packet.pts = i
                packet.stream = data_stream
                container.mux(packet)

        with av.open(output_path) as newcontainer:
            data_stream = newcontainer.streams.data[0]
            klv_stream = newcontainer.streams.data[1]
            assert data_stream.type == "data"
            assert klv_stream.type == "data"
            assert "bin_data" in str(data_stream)
            assert "klv" in str(klv_stream)
            assert data_stream.name == "bin_data"
            assert klv_stream.name == "klv"
        try:
            os.remove(output_path)
        except Exception:
            pass

    def test_attachment_stream(self) -> None:
        input_path = av.datasets.curated(
            "pexels/time-lapse-video-of-night-sky-857195.mp4"
        )
        input_ = av.open(input_path)
        out1_path = "video_with_attachment.mkv"

        with av.open(out1_path, "w") as out1:
            out1.add_attachment(
                name="attachment.txt", mimetype="text/plain", data=b"hello\n"
            )

            in_v = input_.streams.video[0]
            out_v = out1.add_stream_from_template(in_v)

            for packet in input_.demux(in_v):
                if packet.size == 0:
                    continue
                packet.stream = out_v
                out1.mux(packet)

        input_.close()

        with av.open(out1_path) as c:
            attachments = c.streams.attachments
            assert len(attachments) == 1
            att = attachments[0]
            assert att.name == "attachment.txt"
            assert att.mimetype == "text/plain"
            assert att.data == b"hello\n"

        out2_path = "remuxed_attachment.mkv"
        with av.open(out1_path) as ic, av.open(out2_path, "w") as oc:
            stream_map = {}
            for s in ic.streams:
                stream_map[s.index] = oc.add_stream_from_template(s)

            for packet in ic.demux(ic.streams.video):
                if packet.size == 0:
                    continue
                updated_stream = stream_map.get(packet.stream.index)
                if isinstance(updated_stream, av.video.stream.VideoStream):
                    packet.stream = updated_stream
                else:
                    raise ValueError("Expected a VideoStream")
                oc.mux(packet)

        with av.open(out2_path) as c:
            attachments = c.streams.attachments
            assert len(attachments) == 1
            att = attachments[0]
            assert att.name == "attachment.txt"
            assert att.mimetype == "text/plain"
            assert att.data == b"hello\n"

    def test_unknown_stream_type(self) -> None:
        ts_data = _make_unknown_stream_ts()

        with av.open(io.BytesIO(ts_data), format="mpegts") as container:
            assert len(container.streams) == 1
            stream = container.streams[0]
            assert stream.type == "unknown"
            assert type(stream) is av.stream.Stream



import errno
from fractions import Fraction

import numpy as np

import av
from av import AudioFrame, VideoFrame
from av.audio.frame import format_dtypes
from av.filter import Filter, Graph

from .common import TestCase, has_pillow


def generate_audio_frame(
    frame_num: int,
    input_format: str = "s16",
    layout: str = "stereo",
    sample_rate: int = 44100,
    frame_size: int = 1024,
) -> AudioFrame:
    """
    Generate audio frame representing part of the sinusoidal wave
    """
    frame = AudioFrame(format=input_format, layout=layout, samples=frame_size)
    frame.sample_rate = sample_rate
    frame.pts = frame_num * frame_size

    for i in range(frame.layout.nb_channels):
        data = np.zeros(frame_size, dtype=format_dtypes[input_format])
        for j in range(frame_size):
            data[j] = np.sin(2 * np.pi * (frame_num + j) * (i + 1) / float(frame_size))
        frame.planes[i].update(data)  # type: ignore

    return frame


def pull_until_blocked(graph: Graph) -> list[av.VideoFrame]:
    frames: list[av.VideoFrame] = []
    while True:
        try:
            frames.append(graph.vpull())
        except av.FFmpegError as e:
            if e.errno != errno.EAGAIN:
                raise
            return frames


class TestFilters(TestCase):
    def test_filter_descriptor(self) -> None:
        f = Filter("testsrc")
        assert f.name == "testsrc"
        assert f.description == "Generate test pattern."

    def test_generator_graph(self):
        graph = Graph()
        src = graph.add("testsrc")
        lutrgb = graph.add(
            "lutrgb",
            "r=maxval+minval-val:g=maxval+minval-val:b=maxval+minval-val",
            name="invert",
        )
        sink = graph.add("buffersink")
        src.link_to(lutrgb)
        lutrgb.link_to(sink)

        # pads and links
        assert src.outputs[0].link.output is lutrgb.inputs[0]
        assert lutrgb.inputs[0].link.input is src.outputs[0]

        frame = sink.pull()
        assert isinstance(frame, VideoFrame)

        if has_pillow:
            frame.to_image().save(self.sandboxed("mandelbrot2.png"))

    def test_auto_find_sink(self) -> None:
        graph = Graph()
        src = graph.add("testsrc")
        src.link_to(graph.add("buffersink"))
        graph.configure()

        frame = graph.vpull()

        if has_pillow:
            frame.to_image().save(self.sandboxed("mandelbrot3.png"))

    def test_delegate_sink(self) -> None:
        graph = Graph()
        src = graph.add("testsrc")
        src.link_to(graph.add("buffersink"))
        graph.configure()

        frame = src.pull()
        assert isinstance(frame, av.VideoFrame)

        if has_pillow:
            frame.to_image().save(self.sandboxed("mandelbrot4.png"))

    def test_audio_buffer_sink(self):
        graph = Graph()
        audio_buffer = graph.add_abuffer(
            format="fltp",
            sample_rate=48000,
            layout="stereo",
            time_base=Fraction(1, 48000),
        )
        audio_buffer.link_to(graph.add("abuffersink"))
        graph.configure()

        try:
            graph.pull()
        except OSError as e:
            # we haven't pushed any input so expect no frames / EAGAIN
            if e.errno != errno.EAGAIN:
                raise

    def test_audio_buffer_resample(self) -> None:
        graph = Graph()
        graph.link_nodes(
            graph.add_abuffer(
                format="fltp",
                sample_rate=48000,
                layout="stereo",
                time_base=Fraction(1, 48000),
            ),
            graph.add(
                "aformat", "sample_fmts=s16:sample_rates=44100:channel_layouts=stereo"
            ),
            graph.add("abuffersink"),
        ).configure()

        graph.push(
            generate_audio_frame(
                0, input_format="fltp", layout="stereo", sample_rate=48000
            )
        )
        out_frame = graph.pull()
        assert isinstance(out_frame, av.AudioFrame)
        assert out_frame.format.name == "s16"
        assert out_frame.layout.name == "stereo"
        assert out_frame.sample_rate == 44100

    def test_audio_buffer_frame_size(self):
        graph = Graph()
        graph.link_nodes(
            graph.add_abuffer(
                format="fltp",
                sample_rate=48000,
                layout="stereo",
                time_base=Fraction(1, 48000),
            ),
            graph.add("abuffersink"),
        ).configure()
        graph.set_audio_frame_size(256)
        graph.push(
            generate_audio_frame(
                0,
                input_format="fltp",
                layout="stereo",
                sample_rate=48000,
                frame_size=1024,
            )
        )
        out_frame = graph.pull()
        assert out_frame.sample_rate == 48000
        assert out_frame.samples == 256

    def test_audio_buffer_volume_filter(self):
        graph = Graph()
        graph.link_nodes(
            graph.add_abuffer(
                format="fltp",
                sample_rate=48000,
                layout="stereo",
                time_base=Fraction(1, 48000),
            ),
            graph.add("volume", volume="0.5"),
            graph.add("abuffersink"),
        ).configure()

        input_frame = generate_audio_frame(
            0, input_format="fltp", layout="stereo", sample_rate=48000
        )
        graph.push(input_frame)

        out_frame = graph.pull()
        assert out_frame.format.name == "fltp"
        assert out_frame.layout.name == "stereo"
        assert out_frame.sample_rate == 48000

        input_data = input_frame.to_ndarray()
        output_data = out_frame.to_ndarray()

        assert np.allclose(input_data * 0.5, output_data)

    def _test_video_buffer(self, graph):
        input_container = av.open(format="lavfi", file="color=c=pink:duration=1:r=30")
        input_video_stream = input_container.streams.video[0]

        buffer = graph.add_buffer(template=input_video_stream)
        bwdif = graph.add("bwdif", "send_field:tff:all")
        buffersink = graph.add("buffersink")
        buffer.link_to(bwdif)
        bwdif.link_to(buffersink)
        graph.configure()

        for frame in input_container.decode():
            assert frame.time_base == Fraction(1, 30)
            graph.vpush(frame)
            filtered_frames = pull_until_blocked(graph)

            if frame.pts == 0:
                # no output for the first input frame
                assert len(filtered_frames) == 0
            else:
                # we expect two filtered frames per input frame
                assert len(filtered_frames) == 2

                assert filtered_frames[0].pts == (frame.pts - 1) * 2
                assert filtered_frames[0].time_base == Fraction(1, 60)

                assert filtered_frames[1].pts == (frame.pts - 1) * 2 + 1
                assert filtered_frames[1].time_base == Fraction(1, 60)

    def test_video_buffer(self):
        self._test_video_buffer(av.filter.Graph())

    def test_video_buffer_threading(self):
        graph = av.filter.Graph()
        graph.threads = 4
        self._test_video_buffer(graph)

    def test_EOF(self) -> None:
        input_container = av.open(format="lavfi", file="color=c=pink:duration=1:r=30")
        video_stream = input_container.streams.video[0]

        graph = av.filter.Graph()
        video_in = graph.add_buffer(template=video_stream)
        palette_gen_filter = graph.add("palettegen")
        video_out = graph.add("buffersink")
        video_in.link_to(palette_gen_filter)
        palette_gen_filter.link_to(video_out)
        graph.configure()

        for frame in input_container.decode(video=0):
            graph.vpush(frame)

        graph.vpush(None)

        # if we do not push None, we get a BlockingIOError
        palette_frame = graph.vpull()

        assert isinstance(palette_frame, av.VideoFrame)
        assert palette_frame.width == 16
        assert palette_frame.height == 16

    def test_push_at_index(self) -> None:
        # overlay has two video buffer sources; `at` targets a single one,
        # instead of broadcasting the same frame to both (like auto-editor's
        # pushIdx/flushIdx).
        width, height = 16, 16

        base = VideoFrame(width, height, "yuv420p")
        for plane in base.planes:
            plane.update(bytes(plane.buffer_size))
        base.pts = 0
        base.time_base = Fraction(1, 30)

        top = VideoFrame(width, height, "yuv420p")
        for i, plane in enumerate(top.planes):
            plane.update(bytes([200 if i == 0 else 128]) * plane.buffer_size)
        top.pts = 0
        top.time_base = Fraction(1, 30)

        graph = Graph()
        b0 = graph.add_buffer(
            width=width, height=height, format=base.format, time_base=base.time_base
        )
        b1 = graph.add_buffer(
            width=width, height=height, format=top.format, time_base=top.time_base
        )
        overlay = graph.add("overlay", "x=0:y=0")
        sink = graph.add("buffersink")
        b0.link_to(overlay, 0, 0)
        b1.link_to(overlay, 0, 1)
        overlay.link_to(sink)
        graph.configure()

        graph.push(base, at=0)
        graph.push(top, at=1)
        graph.push(None, at=0)
        graph.push(None, at=1)

        out = graph.vpull()
        assert isinstance(out, av.VideoFrame)
        assert (out.width, out.height) == (width, height)

        with self.assertRaises(IndexError):
            graph.push(base, at=2)

    def test_graph_threads(self) -> None:
        graph = Graph()
        assert graph.threads == 0

        graph.threads = 4
        assert graph.threads == 4

        graph.add("testsrc")

        with self.assertRaises(RuntimeError):
            graph.threads = 2



cdef extern from "libavfilter/avfilter.h" nogil:
    cdef int avfilter_version()
    cdef char* avfilter_configuration()
    cdef char* avfilter_license()

    cdef struct AVFilterPad:
        pass

    const char* avfilter_pad_get_name(const AVFilterPad *pads, int index)
    AVMediaType avfilter_pad_get_type(const AVFilterPad *pads, int index)

    cdef unsigned avfilter_filter_pad_count(const AVFilter *filter, int is_output)

    cdef struct AVFilter:
        const char *name
        const char *description
        const AVFilterPad *inputs
        const AVFilterPad *outputs
        const AVClass *priv_class
        int flags

    cdef AVFilter* avfilter_get_by_name(const char *name)
    cdef const AVFilter* av_filter_iterate(void **opaque)

    cdef struct AVFilterContext:
        AVClass *av_class
        AVFilter *filter

        char *name

        unsigned int nb_inputs
        AVFilterPad *input_pads
        AVFilterLink **inputs

        unsigned int nb_outputs
        AVFilterPad *output_pads
        AVFilterLink **outputs

    cdef int avfilter_init_str(AVFilterContext *ctx, const char *args)
    cdef int avfilter_init_dict(AVFilterContext *ctx, AVDictionary **options)
    cdef void avfilter_free(AVFilterContext*)
    cdef AVClass* avfilter_get_class()

    cdef struct AVFilterLink:
        AVFilterContext *src
        AVFilterPad *srcpad
        AVFilterContext *dst
        AVFilterPad *dstpad
        AVMediaType Type
        int w
        int h
        AVRational sample_aspect_ratio
        uint64_t channel_layout
        int sample_rate
        int format
        AVRational time_base

    cdef struct AVFilterGraph:
        int nb_filters
        AVFilterContext **filters
        int nb_threads

    cdef struct AVFilterInOut:
        char *name
        AVFilterContext *filter_ctx
        int pad_idx
        AVFilterInOut *next

    cdef AVFilterGraph* avfilter_graph_alloc()
    cdef void avfilter_graph_free(AVFilterGraph **ptr)
    cdef AVFilterContext* avfilter_graph_alloc_filter(
        AVFilterGraph *graph,
        const AVFilter *filter,
        const char *name
    )
    cdef int avfilter_graph_create_filter(
        AVFilterContext **filt_ctx,
        AVFilter *filt,
        const char *name,
        const char *args,
        void *opaque,
        AVFilterGraph *graph_ctx
    )
    cdef int avfilter_link(
        AVFilterContext *src,
        unsigned int srcpad,
        AVFilterContext *dst,
        unsigned int dstpad
    )
    cdef int avfilter_graph_config(AVFilterGraph *graph, void *logctx)
    int avfilter_process_command(
        AVFilterContext *filter, const char *cmd, const char *arg, char *res,
        int res_len, int flags,
    )

cdef extern from "libavfilter/buffersink.h" nogil:
    cdef void av_buffersink_set_frame_size(AVFilterContext *ctx, unsigned frame_size)
    int av_buffersink_get_frame(AVFilterContext *ctx, AVFrame *frame)

cdef extern from "libavfilter/buffersrc.h" nogil:
    int av_buffersrc_write_frame(AVFilterContext *ctx, const AVFrame *frame)



import av

av.logging.set_level(av.logging.VERBOSE)

input_file = av.open("input.wav")
output_file = av.open("output.wav", mode="w")

input_stream = input_file.streams.audio[0]
output_stream = output_file.add_stream("pcm_s16le", rate=input_stream.rate)

graph = av.filter.Graph()
graph.link_nodes(
    graph.add_abuffer(template=input_stream),
    graph.add("atempo", "2.0"),
    graph.add("abuffersink"),
).configure()

for frame in input_file.decode(input_stream):
    graph.push(frame)
    while True:
        try:
            for packet in output_stream.encode(graph.pull()):
                output_file.mux(packet)
        except (av.BlockingIOError, av.EOFError):
            break

# Flush the stream
for packet in output_stream.encode(None):
    output_file.mux(packet)

input_file.close()
output_file.close()



import pytest

from av import AudioFormat, Codec, VideoFormat, codecs_available
from av.codec import PixFmtLoss, find_best_pix_fmt_of_list
from av.codec.codec import UnknownCodecError


def test_codec_bogus() -> None:
    with pytest.raises(UnknownCodecError):
        Codec("bogus123")
    with pytest.raises(UnknownCodecError):
        Codec("bogus123", "w")


def test_codec_mpeg4_decoder() -> None:
    c = Codec("mpeg4")

    assert c.name == "mpeg4"
    assert c.long_name == "MPEG-4 part 2"
    assert c.type == "video"
    assert c.id in (12, 13)
    assert c.is_decoder
    assert not c.is_encoder
    assert c.delay

    assert c.audio_formats is None and c.audio_rates is None

    # formats = c.video_formats
    # assert formats
    # assert isinstance(formats[0], VideoFormat)
    # assert any(f.name == "yuv420p" for f in formats)

    assert c.frame_rates is None


def test_codec_mpeg4_encoder() -> None:
    c = Codec("mpeg4", "w")
    assert c.name == "mpeg4"
    assert c.long_name == "MPEG-4 part 2"
    assert c.type == "video"
    assert c.id in (12, 13)
    assert c.is_encoder
    assert not c.is_decoder
    assert c.delay

    assert c.audio_formats is None and c.audio_rates is None

    formats = c.video_formats
    assert formats
    assert isinstance(formats[0], VideoFormat)
    assert any(f.name == "yuv420p" for f in formats)
    assert c.frame_rates is None


def test_codec_opus_decoder() -> None:
    c = Codec("opus")

    assert c.name == "opus"
    assert c.long_name == "Opus"
    assert c.type == "audio"
    assert c.is_decoder
    assert not c.is_encoder
    assert c.delay

    assert c.audio_formats is None and c.audio_rates is None
    assert c.video_formats is None and c.frame_rates is None


def test_codec_opus_encoder() -> None:
    c = Codec("opus", "w")
    assert c.name in ("opus", "libopus")
    assert c.canonical_name == "opus"
    assert c.long_name in ("Opus", "libopus Opus")
    assert c.type == "audio"
    assert c.is_encoder
    assert not c.is_decoder
    assert c.delay

    # audio
    formats = c.audio_formats
    assert formats
    assert isinstance(formats[0], AudioFormat)
    assert any(f.name in ("flt", "fltp") for f in formats)

    assert c.audio_rates is not None
    assert 48000 in c.audio_rates

    assert c.video_formats is None and c.frame_rates is None


def test_codecs_available() -> None:
    assert codecs_available


def test_find_best_pix_fmt_of_list_empty() -> None:
    best, loss = find_best_pix_fmt_of_list([], "rgb24")
    assert best is None
    assert loss == 0
    assert loss is PixFmtLoss.NONE


@pytest.mark.parametrize(
    "pix_fmts,src_pix_fmt,expected_best",
    [
        (["rgb24", "yuv420p"], "rgb24", "rgb24"),
        (["rgb24"], "yuv420p", "rgb24"),
        (["yuv420p"], "rgb24", "yuv420p"),
        ([VideoFormat("yuv420p")], VideoFormat("rgb24"), "yuv420p"),
        (
            ["yuv420p", "yuv444p", "gray", "rgb24", "rgba", "bgra", "yuyv422"],
            "rgba",
            "rgba",
        ),
    ],
)
def test_find_best_pix_fmt_of_list_best(pix_fmts, src_pix_fmt, expected_best) -> None:
    best, loss = find_best_pix_fmt_of_list(pix_fmts, src_pix_fmt)
    assert best is not None
    assert best.name == expected_best
    assert isinstance(loss, int)


@pytest.mark.parametrize(
    "pix_fmts,src_pix_fmt",
    [
        (["__unknown_pix_fmt"], "rgb24"),
        (["rgb24"], "__unknown_pix_fmt"),
    ],
)
def test_find_best_pix_fmt_of_list_unknown_pix_fmt(pix_fmts, src_pix_fmt) -> None:
    with pytest.raises(ValueError, match="not a pixel format"):
        find_best_pix_fmt_of_list(pix_fmts, src_pix_fmt)


@pytest.mark.parametrize(
    "pix_fmts,src_pix_fmt",
    [
        (["rgb24", "bgr24", "gray", "yuv420p", "yuv444p", "yuyv422"], "nv12"),
        (["yuv420p", "yuv444p", "gray", "yuv420p"], "rgb24"),
        (["rgb24", "rgba", "bgra", "rgb24", "gray"], "yuv420p"),
    ],
)
def test_find_best_pix_fmt_of_list_picks_from_list(pix_fmts, src_pix_fmt) -> None:
    best, loss = find_best_pix_fmt_of_list(pix_fmts, src_pix_fmt)
    assert best is not None
    assert best.name in set(pix_fmts)
    assert isinstance(loss, int)


def test_find_best_pix_fmt_of_list_alpha_loss_flagged_when_used() -> None:
    best, loss = find_best_pix_fmt_of_list(["rgb24"], "rgba", has_alpha=True)
    assert best is not None
    assert best.name == "rgb24"
    assert loss != 0
    assert isinstance(loss, PixFmtLoss)
    assert loss & PixFmtLoss.ALPHA


def test_find_best_pix_fmt_of_list_loss_flags() -> None:
    # An identical format loses nothing.
    _, loss = find_best_pix_fmt_of_list(["yuv420p"], "yuv420p")
    assert loss is PixFmtLoss.NONE

    # Converting color to grayscale drops the chroma planes.
    _, loss = find_best_pix_fmt_of_list(["gray"], "rgb24")
    assert isinstance(loss, PixFmtLoss)
    assert loss & PixFmtLoss.CHROMA



import gc
from fractions import Fraction
from typing import cast

import numpy
import pytest

import av
from av import VideoFrame
from av.codec.hwaccel import HWAccel
from av.video.codeccontext import VideoCodecContext

from .common import assertNdarraysEqual, fate_png


def _make_u8(shape: tuple[int, ...]) -> numpy.ndarray:
    return numpy.arange(int(numpy.prod(shape)), dtype=numpy.uint8).reshape(shape)


def _make_u16(shape: tuple[int, ...]) -> numpy.ndarray:
    return numpy.arange(int(numpy.prod(shape)), dtype=numpy.uint16).reshape(shape)


def _plane_to_2d(plane, height: int, width: int, dtype) -> numpy.ndarray:
    itemsize = numpy.dtype(dtype).itemsize
    assert plane.line_size % itemsize == 0
    pitch_elems = plane.line_size // itemsize
    arr = numpy.frombuffer(memoryview(plane), dtype=dtype).reshape(height, pitch_elems)
    return arr[:, :width]


def _get_cuda_backend():
    try:
        import cupy  # type: ignore

        try:
            if cupy.cuda.runtime.getDeviceCount() > 0:
                return ("cupy", cupy)
        except Exception:
            pass
    except Exception:
        pass

    try:
        import torch  # type: ignore

        if torch.cuda.is_available():
            return ("torch", torch)
    except Exception:
        pass

    return None


def test_hwaccel_validation_and_primary_ctx() -> None:
    hw = HWAccel(device_type="cuda")
    assert hw.is_hw_owned == False
    assert "primary_ctx" not in hw.options

    hw = HWAccel(device_type="cuda", is_hw_owned=True)
    assert hw.is_hw_owned == True
    assert hw.options.get("primary_ctx") == "1"

    hw = HWAccel(device_type="cuda", is_hw_owned=True, options={"primary_ctx": "0"})
    assert hw.options.get("primary_ctx") == "0"


def test_video_frame_from_dlpack_nv12_cpu_basic_zero_copy_and_lifetime() -> None:
    width, height = 64, 48
    y = _make_u8((height, width))
    uv = _make_u8((height // 2, width // 2, 2))

    frame = VideoFrame.from_dlpack((y, uv), format="nv12")

    assert frame.format.name == "nv12"
    assert frame.width == width
    assert frame.height == height
    assert len(frame.planes) == 2
    assert frame.planes[0].width == width
    assert frame.planes[0].height == height
    assert frame.planes[1].width == width // 2
    assert frame.planes[1].height == height // 2
    assert frame.planes[0].line_size == width
    assert frame.planes[1].line_size == width

    y_plane = _plane_to_2d(frame.planes[0], height, width, numpy.uint8)
    uv_plane = _plane_to_2d(frame.planes[1], height // 2, width, numpy.uint8)
    assertNdarraysEqual(y_plane, y)
    assertNdarraysEqual(uv_plane, uv.reshape(height // 2, width))

    y[0, 0] = 123
    uv[0, 0, 0] = 11
    uv[0, 0, 1] = 22

    expected_y_bytes = y.tobytes()
    expected_uv_bytes = uv.reshape(height // 2, width).tobytes()

    assert memoryview(frame.planes[0])[0] == 123
    assert memoryview(frame.planes[1])[0] == 11
    assert memoryview(frame.planes[1])[1] == 22

    del y
    del uv
    gc.collect()

    assert bytes(frame.planes[0]) == expected_y_bytes
    assert bytes(frame.planes[1]) == expected_uv_bytes


def test_video_frame_from_dlpack_nv12_cpu_with_pitch_and_dlpack_export() -> None:
    width, height = 64, 48
    pad = 16

    y_base = _make_u8((height, width + pad))
    y = y_base[:, :width]
    uv_base = _make_u8((height // 2, (width + pad) // 2, 2))
    uv = uv_base[:, : width // 2, :]

    frame = VideoFrame.from_dlpack((y, uv), format="nv12")

    assert frame.planes[0].line_size == width + pad
    assert frame.planes[1].line_size == width + pad
    assert frame.planes[0].buffer_size == (width + pad) * height
    assert frame.planes[1].buffer_size == (width + pad) * (height // 2)

    y_plane = _plane_to_2d(frame.planes[0], height, width, numpy.uint8)
    uv_plane = _plane_to_2d(frame.planes[1], height // 2, width, numpy.uint8)
    assertNdarraysEqual(y_plane, y)
    assertNdarraysEqual(uv_plane, uv.reshape(height // 2, width))

    assert frame.planes[0].__dlpack_device__() == (1, 0)

    y_dl = numpy.from_dlpack(frame.planes[0])
    uv_dl = numpy.from_dlpack(frame.planes[1])

    assert y_dl.shape == (height, width)
    assert y_dl.dtype == numpy.uint8
    assert y_dl.strides == (width + pad, 1)
    assertNdarraysEqual(y_dl, y)

    assert uv_dl.shape == (height // 2, width // 2, 2)
    assert uv_dl.dtype == numpy.uint8
    assert uv_dl.strides == (width + pad, 2, 1)
    assertNdarraysEqual(uv_dl, uv)

    expected_y = numpy.array(y, copy=True)
    expected_uv = numpy.array(uv, copy=True)

    del frame
    del y
    del uv
    del y_base
    del uv_base
    gc.collect()

    assertNdarraysEqual(y_dl, expected_y)
    assertNdarraysEqual(uv_dl, expected_uv)


def test_video_frame_from_dlpack_nv12_cpu_accepts_uv_2d() -> None:
    width, height = 64, 48
    y = _make_u8((height, width))
    uv2d = _make_u8((height // 2, width))

    frame = VideoFrame.from_dlpack((y, uv2d), format="nv12")

    uv_plane = _plane_to_2d(frame.planes[1], height // 2, width, numpy.uint8)
    assertNdarraysEqual(uv_plane, uv2d)

    uv_dl = numpy.from_dlpack(frame.planes[1])
    assert uv_dl.shape == (height // 2, width // 2, 2)
    assertNdarraysEqual(uv_dl, uv2d.reshape(height // 2, width // 2, 2))


def test_video_frame_from_dlpack_accepts_video_plane_objects() -> None:
    width, height = 64, 48
    y = _make_u8((height, width))
    uv = _make_u8((height // 2, width // 2, 2))

    frame1 = VideoFrame.from_dlpack((y, uv), format="nv12")
    frame2 = VideoFrame.from_dlpack((frame1.planes[0], frame1.planes[1]), format="nv12")

    assert bytes(frame2.planes[0]) == bytes(frame1.planes[0])
    assert bytes(frame2.planes[1]) == bytes(frame1.planes[1])


@pytest.mark.parametrize("fmt", ["p010le", "p016le"])
def test_video_frame_from_dlpack_p010_p016_cpu(fmt: str) -> None:
    width, height = 64, 48
    y = _make_u16((height, width))
    uv = _make_u16((height // 2, width // 2, 2))

    frame = VideoFrame.from_dlpack((y, uv), format=fmt)

    assert frame.format.name == fmt
    assert len(frame.planes) == 2
    assert frame.planes[0].line_size == width * 2
    assert frame.planes[1].line_size == width * 2

    y_plane = _plane_to_2d(frame.planes[0], height, width, numpy.uint16)
    uv_plane = _plane_to_2d(frame.planes[1], height // 2, width, numpy.uint16)
    assertNdarraysEqual(y_plane, y)
    assertNdarraysEqual(uv_plane, uv.reshape(height // 2, width))

    y_dl = numpy.from_dlpack(frame.planes[0])
    uv_dl = numpy.from_dlpack(frame.planes[1])

    assert y_dl.dtype == numpy.uint16
    assert y_dl.shape == (height, width)
    assert y_dl.strides == (width * 2, 2)
    assertNdarraysEqual(y_dl, y)

    assert uv_dl.dtype == numpy.uint16
    assert uv_dl.shape == (height // 2, width // 2, 2)
    assert uv_dl.strides == (width * 2, 4, 2)
    assertNdarraysEqual(uv_dl, uv)


def test_video_plane_dlpack_export_keeps_frame_alive_after_gc() -> None:
    container = av.open(fate_png())
    frame = next(container.decode(video=0))
    frame_nv12 = frame.reformat(format="nv12")

    width = frame_nv12.width
    height = frame_nv12.height
    line_size = frame_nv12.planes[0].line_size
    expected = _plane_to_2d(frame_nv12.planes[0], height, width, numpy.uint8).copy()

    y_dl = numpy.from_dlpack(frame_nv12.planes[0])
    assert y_dl.shape == (height, width)
    assert y_dl.strides == (line_size, 1)

    del frame_nv12
    del frame
    del container
    gc.collect()

    assertNdarraysEqual(y_dl, expected)


def test_video_plane_dlpack_export_packed_rgb_cpu() -> None:
    # Packed formats interleave several components in one plane and export as
    # a 3D (H, W, C) tensor (issue #2217).
    rgb = (numpy.arange(16 * 24 * 3) % 251).astype(numpy.uint8).reshape(16, 24, 3)
    frame = VideoFrame.from_ndarray(rgb, format="rgb24")
    plane = frame.planes[0]
    assert plane.__dlpack_device__() == (1, 0)

    arr = numpy.from_dlpack(plane)
    assert arr.shape == (16, 24, 3)
    assert arr.strides == (plane.line_size, 3, 1)
    assert arr.dtype == numpy.uint8
    assertNdarraysEqual(arr, rgb)


def test_video_plane_dlpack_export_planar_yuv_cpu() -> None:
    # Planar formats expose each single-component plane as a 2D (H, W) tensor
    # (issue #2217).
    frame = VideoFrame(16, 16, "yuv420p")
    for index, (h, w) in enumerate([(16, 16), (8, 8), (8, 8)]):
        plane = frame.planes[index]
        assert plane.__dlpack_device__() == (1, 0)
        arr = numpy.from_dlpack(plane)
        assert arr.shape == (h, w)
        assert arr.strides == (plane.line_size, 1)
        assert arr.dtype == numpy.uint8


def test_video_plane_dlpack_export_planar_yuv16_cpu() -> None:
    # 16-bit planar formats export as uint16.
    frame = VideoFrame(16, 16, "yuv420p10le")
    arr = numpy.from_dlpack(frame.planes[0])
    assert arr.shape == (16, 16)
    assert arr.dtype == numpy.uint16


def test_video_plane_dlpack_unsupported_format_raises() -> None:
    # Palette formats still cannot be exported.
    frame = VideoFrame(16, 16, "pal8")
    assert frame.planes[0].__dlpack_device__() == (1, 0)

    with pytest.raises(
        NotImplementedError,
        match="bitstream, palette, and Bayer formats are not supported",
    ):
        frame.planes[0].__dlpack__()


def test_sw_format_none_for_software_frame() -> None:
    assert VideoFrame(16, 16, "yuv420p").sw_format is None


def test_video_frame_from_dlpack_requires_two_planes() -> None:
    y = numpy.zeros((4, 4), dtype=numpy.uint8)
    with pytest.raises(ValueError, match="2-plane"):
        VideoFrame.from_dlpack(y, format="nv12")


def test_video_frame_from_dlpack_rejects_packed_format() -> None:
    # Packed formats interleave several components in one plane and cannot be
    # imported (only planar single-component formats and the nv12 family are).
    rgb = numpy.zeros((16, 16, 3), dtype=numpy.uint8)

    with pytest.raises(NotImplementedError, match="is not supported"):
        VideoFrame.from_dlpack((rgb,), format="rgb24")


@pytest.mark.parametrize(
    "fmt,dtype,planes_hw",
    [
        ("yuv420p", numpy.uint8, [(48, 64), (24, 32), (24, 32)]),
        ("yuv422p", numpy.uint8, [(48, 64), (48, 32), (48, 32)]),
        ("yuv444p", numpy.uint8, [(48, 64), (48, 64), (48, 64)]),
        ("gray", numpy.uint8, [(48, 64)]),
        ("yuv420p10le", numpy.uint16, [(48, 64), (24, 32), (24, 32)]),
    ],
)
def test_video_frame_from_dlpack_planar_cpu(fmt, dtype, planes_hw) -> None:
    # Issue #2217: planar formats whose planes each hold one component round
    # trip through DLPack without a NumPy intermediate.
    width, height = 64, 48
    make = _make_u16 if dtype == numpy.uint16 else _make_u8
    src = [make((h, w)) for (h, w) in planes_hw]

    frame = VideoFrame.from_dlpack(tuple(src), format=fmt)

    assert frame.format.name == fmt
    assert frame.width == width and frame.height == height
    assert len(frame.planes) == len(src)

    for i, plane in enumerate(frame.planes):
        arr = numpy.from_dlpack(plane)
        assert arr.dtype == dtype
        assert arr.shape == planes_hw[i]
        assertNdarraysEqual(arr, src[i])


def test_video_frame_from_dlpack_yuv420p_zero_copy_and_lifetime() -> None:
    width, height = 64, 48
    y = _make_u8((height, width))
    u = _make_u8((height // 2, width // 2))
    v = _make_u8((height // 2, width // 2))

    frame = VideoFrame.from_dlpack((y, u, v), format="yuv420p")

    # Mutating the source is visible through the frame (zero copy).
    y[0, 0] = 200
    assert memoryview(frame.planes[0])[0] == 200

    expected = [y.copy(), u.copy(), v.copy()]
    del y, u, v
    gc.collect()

    for i, plane in enumerate(frame.planes):
        assertNdarraysEqual(numpy.from_dlpack(plane), expected[i])


def test_video_frame_from_dlpack_yuv420p_with_pitch() -> None:
    width, height = 64, 48
    pad = 16

    y = _make_u8((height, width + pad))[:, :width]
    u = _make_u8((height // 2, (width + pad) // 2))[:, : width // 2]
    v = _make_u8((height // 2, (width + pad) // 2))[:, : width // 2]

    frame = VideoFrame.from_dlpack((y, u, v), format="yuv420p")

    assert frame.planes[0].line_size == width + pad
    assert frame.planes[1].line_size == (width + pad) // 2
    assertNdarraysEqual(numpy.from_dlpack(frame.planes[0]), y)
    assertNdarraysEqual(numpy.from_dlpack(frame.planes[1]), u)
    assertNdarraysEqual(numpy.from_dlpack(frame.planes[2]), v)


def test_video_frame_from_dlpack_planar_wrong_plane_count() -> None:
    y = numpy.zeros((48, 64), dtype=numpy.uint8)
    u = numpy.zeros((24, 32), dtype=numpy.uint8)

    with pytest.raises(ValueError, match=r"requires 3 plane\(s\), got 2"):
        VideoFrame.from_dlpack((y, u), format="yuv420p")


def test_video_frame_from_dlpack_planar_rejects_odd_dimensions() -> None:
    y = numpy.zeros((48, 63), dtype=numpy.uint8)
    u = numpy.zeros((24, 32), dtype=numpy.uint8)
    v = numpy.zeros((24, 32), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="must be even"):
        VideoFrame.from_dlpack((y, u, v), format="yuv420p")


def test_video_frame_from_dlpack_planar_rejects_palette() -> None:
    idx = numpy.zeros((16, 16), dtype=numpy.uint8)

    with pytest.raises(NotImplementedError, match="palette"):
        VideoFrame.from_dlpack((idx,), format="pal8")


def test_video_frame_from_dlpack_rejects_device_id_for_cpu() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="device_id must be 0 for CPU tensors"):
        VideoFrame.from_dlpack((y, uv), format="nv12", device_id=1)


def test_video_frame_from_dlpack_requires_both_width_height_or_neither() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="either specify both width/height or neither"):
        VideoFrame.from_dlpack((y, uv), format="nv12", width=width, height=0)


def test_video_frame_from_dlpack_rejects_plane0_shape_mismatch_with_width_height() -> (
    None
):
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="plane 0 shape does not match width/height"):
        VideoFrame.from_dlpack((y, uv), format="nv12", width=width + 2, height=height)


def test_video_frame_from_dlpack_rejects_odd_dimensions() -> None:
    width, height = 63, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="width/height must be even"):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_rejects_noncontiguous_plane0_last_dim() -> None:
    width, height = 64, 48
    y_full = numpy.zeros((height, width * 2), dtype=numpy.uint8)
    y = y_full[:, ::2]
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    with pytest.raises(
        ValueError, match="plane 0 must be contiguous in the last dimension"
    ):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_rejects_noncontiguous_uv_plane_last_dim_2d() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv_full = numpy.zeros((height // 2, width * 2), dtype=numpy.uint8)
    uv = uv_full[:, ::2]

    with pytest.raises(
        ValueError, match="plane 1 must be contiguous in the last dimension"
    ):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_rejects_unexpected_uv_strides_3d() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv_full = numpy.zeros((height // 2, width // 2, 4), dtype=numpy.uint8)
    uv = uv_full[:, :, :2]

    with pytest.raises(ValueError, match="unexpected UV plane strides"):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_rejects_wrong_dtype_plane0() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint16)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    with pytest.raises(TypeError, match="unexpected dtype for plane 0"):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_rejects_wrong_dtype_plane1() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint16)

    with pytest.raises(TypeError, match="unexpected dtype for plane 1"):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_p010le_requires_uint16() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    with pytest.raises(TypeError, match="unexpected dtype for plane 0"):
        VideoFrame.from_dlpack((y, uv), format="p010le")


def test_video_frame_from_dlpack_rejects_plane0_ndim_not_2() -> None:
    y = numpy.zeros((4, 4, 1), dtype=numpy.uint8)
    uv = numpy.zeros((2, 4), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="plane 0 must be 2D"):
        VideoFrame.from_dlpack((y, uv), format="nv12", width=4, height=4)


def test_video_frame_from_dlpack_rejects_plane1_ndim_not_2_or_3() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width, 1, 1), dtype=numpy.uint8)

    with pytest.raises(ValueError, match="plane 1 must be 2D or 3D"):
        VideoFrame.from_dlpack((y, uv), format="nv12")


def test_video_frame_from_dlpack_reusing_capsule_raises_typeerror() -> None:
    width, height = 64, 48
    y = numpy.zeros((height, width), dtype=numpy.uint8)
    uv = numpy.zeros((height // 2, width // 2, 2), dtype=numpy.uint8)

    cap0 = y.__dlpack__()
    cap1 = uv.__dlpack__()

    VideoFrame.from_dlpack((cap0, cap1), format="nv12", width=width, height=height)

    with pytest.raises(TypeError, match="expected a DLPack capsule"):
        VideoFrame.from_dlpack((cap0, cap1), format="nv12", width=width, height=height)


def test_video_frame_from_dlpack_invalid_plane_object_raises_typeerror() -> None:
    with pytest.raises(TypeError, match="expected a DLPack capsule"):
        VideoFrame.from_dlpack((object(), object()), format="nv12", width=64, height=48)


def test_video_frame_from_dlpack_cuda_hw_frame_behavior_if_available() -> None:
    backend = _get_cuda_backend()
    if backend is None:
        pytest.skip("CUDA backend (cupy/torch) not available.")

    width, height = 64, 48
    name, mod = backend

    try:
        if name == "cupy":
            try:
                ndev = int(mod.cuda.runtime.getDeviceCount())
            except Exception:
                ndev = 1

            device_id = 1 if ndev > 1 else 0
            with mod.cuda.Device(device_id):
                y = mod.arange(height * width, dtype=mod.uint8).reshape(height, width)
                uv = mod.arange(
                    (height // 2) * (width // 2) * 2, dtype=mod.uint8
                ).reshape(height // 2, width // 2, 2)
                expected_device = y.__dlpack_device__()
                frame = VideoFrame.from_dlpack((y, uv), format="nv12")

                assert frame.format.name == "cuda"
                assert len(frame.planes) == 2

                with pytest.raises(
                    TypeError, match="Hardware frame planes do not support"
                ):
                    memoryview(frame.planes[0])

                assert frame.planes[0].__dlpack_device__() == expected_device

                cap_y = frame.planes[0].__dlpack__()
                if hasattr(mod, "fromDlpack"):
                    y2 = mod.fromDlpack(cap_y)
                else:
                    y2 = mod.from_dlpack(cap_y)

                assert y2.shape == y.shape
                assert mod.all(y2 == y).item()

                with pytest.raises(
                    ValueError,
                    match="Cannot convert a hardware frame to numpy directly",
                ):
                    frame.to_ndarray(format="cuda")

        else:
            try:
                ndev = int(mod.cuda.device_count())
            except Exception:
                ndev = 1

            device_id = 1 if ndev > 1 else 0
            device = f"cuda:{device_id}"

            y = mod.arange(height * width, device=device, dtype=mod.uint8).reshape(
                height, width
            )
            uv = mod.arange(
                (height // 2) * (width // 2) * 2, device=device, dtype=mod.uint8
            ).reshape(height // 2, width // 2, 2)

            expected_device = y.__dlpack_device__()
            frame = VideoFrame.from_dlpack((y, uv), format="nv12")

            assert frame.format.name == "cuda"
            assert len(frame.planes) == 2

            with pytest.raises(TypeError, match="Hardware frame planes do not support"):
                memoryview(frame.planes[0])

            assert frame.planes[0].__dlpack_device__() == expected_device

            cap_y = frame.planes[0].__dlpack__()
            y2 = mod.utils.dlpack.from_dlpack(cap_y)

            assert tuple(y2.shape) == tuple(y.shape)
            assert mod.equal(y2, y)

            with pytest.raises(
                ValueError, match="Cannot convert a hardware frame to numpy directly"
            ):
                frame.to_ndarray(format="cuda")
    except av.FFmpegError as e:
        pytest.skip(f"CUDA hwcontext not available in this build/runtime: {e}")


def test_encode_cuda_frame_with_nvenc_if_available() -> None:
    # Issue #2199: a CUDA frame from DLPack should encode on the GPU directly.
    # Its hw_frames_ctx must propagate to the encoder before avcodec_open2.
    backend = _get_cuda_backend()
    if backend is None:
        pytest.skip("CUDA backend (cupy/torch) not available.")

    name, mod = backend
    width, height = 256, 256

    try:
        if name == "torch":
            y = mod.zeros((height, width), dtype=mod.uint8, device="cuda")
            uv = mod.zeros((height // 2, width // 2, 2), dtype=mod.uint8, device="cuda")
        else:
            y = mod.zeros((height, width), dtype=mod.uint8)
            uv = mod.zeros((height // 2, width // 2, 2), dtype=mod.uint8)

        frame = VideoFrame.from_dlpack((y, uv), format="nv12")
        assert frame.format.name == "cuda"
        assert frame.sw_format is not None and frame.sw_format.name == "nv12"

        cc = cast(VideoCodecContext, av.CodecContext.create("h264_nvenc", "w"))
        cc.width = width
        cc.height = height
        cc.time_base = Fraction(1, 24)
        cc.framerate = Fraction(24, 1)
        cc.pix_fmt = "cuda"

        packets = cc.encode(frame)
        packets += cc.encode(None)  # flush
        assert any(p.size for p in packets)
    except av.FFmpegError as e:
        pytest.skip(f"nvenc/CUDA not available in this build/runtime: {e}")



from __future__ import annotations

import io
import struct
from typing import cast

import numpy as np
import pytest

import av
from av.sidedata.sidedata import SideData
from av.video.stream import VideoStream

WIDTH = 320
HEIGHT = 240
DURATION = 10

# The 8 EXIF orientations as 3x3 transformation matrices, built the same way as
# the application code: a 90 deg rotation generator (R) and a horizontal-flip
# generator (F). Orientations 2, 4, 5, 7 are reflections, which a scalar
# rotation cannot represent -- so these are verified by comparing the full
# matrix that round-trips through the container.
_R = np.asarray([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=float)  # exif 8
_F = np.asarray([[-1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float)  # exif 2

EXIF_MATRICES = {
    1: np.eye(3),
    2: _F,
    3: _R @ _R,
    4: _F @ _R @ _R,
    5: _F @ _R @ _R @ _R,
    6: _R @ _R @ _R,
    7: _F @ _R,
    8: _R,
}

# Pure-rotation orientations also have a well-defined scalar rotation, reported
# by av_display_rotation_get() (counter-clockwise, range [-180, 180]).
EXPECTED_ROTATION = {1: 0, 3: 180, 6: -90, 8: 90}

# Each EXIF orientation expressed through the convenience API as
# (degrees_ccw, hflip, vflip). Verified to reproduce EXIF_MATRICES exactly.
EXIF_VIA_ROTATION = {
    1: (0, False, False),
    2: (0, True, False),
    3: (0, True, True),
    4: (0, False, True),
    5: (90, True, False),
    6: (90, True, True),
    7: (90, False, True),
    8: (90, False, False),
}

# One encoder per codec family we care about, plus the near-universal mpeg4
# baseline. Unavailable encoders are skipped at runtime so the suite stays
# portable across FFmpeg builds.
CODECS = ["mpeg4", "libx264", "libopenh264", "libx265", "libsvtav1", "libaom-av1"]


def matrix_to_ints(matrix: np.ndarray) -> list[int]:
    """Convert a 3x3 matrix to FFmpeg's AV_PKT_DATA_DISPLAYMATRIX integers.

    Layout (a, b, u, c, d, v, x, y, w): 16.16 fixed point everywhere except
    u, v, w (indices 2, 5, 8) which are 2.30.
    """
    flat = [float(v) for v in matrix.reshape(-1)]
    return [
        int(round(v * (1 << 30))) if i in (2, 5, 8) else int(round(v * (1 << 16)))
        for i, v in enumerate(flat)
    ]


def _has_encoder(name: str) -> bool:
    try:
        av.codec.Codec(name, "w")
    except Exception:
        return False
    return True


def _encode(codec_name: str, matrix: list[int] | None) -> io.BytesIO:
    buf = io.BytesIO()
    container = av.open(buf, "w", format="mp4")
    stream = cast(VideoStream, container.add_stream(codec_name, rate=24))
    stream.width = WIDTH
    stream.height = HEIGHT
    stream.pix_fmt = "yuv420p"

    if matrix is not None:
        stream.set_display_matrix(matrix)

    for i in range(DURATION):
        img = np.full((HEIGHT, WIDTH, 3), (i * 8) % 256, dtype=np.uint8)
        frame = av.VideoFrame.from_ndarray(img, format="rgb24")
        for packet in stream.encode(frame):
            container.mux(packet)
    for packet in stream.encode():
        container.mux(packet)
    container.close()

    buf.seek(0)
    return buf


def _read_frame(buf: io.BytesIO) -> av.VideoFrame:
    with av.open(buf, "r", format="mp4") as container:
        return next(container.decode(video=0))


def _read_matrix(frame: av.VideoFrame) -> list[int] | None:
    sd = frame.side_data.get("DISPLAYMATRIX")
    if sd is None:
        return None
    return list(struct.unpack("=9i", bytes(cast(SideData, sd))))


@pytest.mark.parametrize("codec_name", CODECS)
@pytest.mark.parametrize("orientation", sorted(EXIF_MATRICES))
def test_exif_orientation_roundtrip(codec_name: str, orientation: int) -> None:
    if not _has_encoder(codec_name):
        pytest.skip(f"encoder {codec_name} not available")

    expected = matrix_to_ints(EXIF_MATRICES[orientation])
    frame = _read_frame(_encode(codec_name, expected))
    got = _read_matrix(frame)

    identity = matrix_to_ints(np.eye(3))
    if expected == identity:
        # Identity is the container default; demuxers emit no side data for it.
        assert got is None
        assert frame.rotation == 0
    else:
        assert got == expected, f"exif {orientation}: wrote {expected}, read {got}"

    if orientation in EXPECTED_ROTATION:
        rotation = frame.rotation
        # 180 may come back negated; rotations are exact otherwise.
        if abs(EXPECTED_ROTATION[orientation]) == 180:
            assert abs(rotation) == 180
        else:
            assert rotation == EXPECTED_ROTATION[orientation]


@pytest.mark.parametrize("degrees,expected", [(0, 0), (90, 90), (180, 180), (270, -90)])
def test_set_display_rotation_roundtrip(degrees: int, expected: int) -> None:
    # The public angle is counter-clockwise, matching VideoFrame.rotation.
    buf = io.BytesIO()
    container = av.open(buf, "w", format="mp4")
    stream = container.add_stream("mpeg4", rate=24)
    stream.width = WIDTH
    stream.height = HEIGHT
    stream.pix_fmt = "yuv420p"
    stream.set_display_rotation(degrees)
    for i in range(DURATION):
        frame = av.VideoFrame.from_ndarray(
            np.zeros((HEIGHT, WIDTH, 3), dtype=np.uint8), format="rgb24"
        )
        for packet in stream.encode(frame):
            container.mux(packet)
    for packet in stream.encode():
        container.mux(packet)
    container.close()

    buf.seek(0)
    rotation = _read_frame(buf).rotation
    if abs(expected) == 180:
        assert abs(rotation) == 180
    else:
        assert rotation == expected


@pytest.mark.parametrize("orientation", sorted(EXIF_VIA_ROTATION))
def test_convenience_reaches_all_exif_orientations(orientation: int) -> None:
    # set_display_rotation(degrees, hflip, vflip) must reproduce the exact same
    # matrix as the explicit EXIF table for every one of the 8 orientations.
    degrees, hflip, vflip = EXIF_VIA_ROTATION[orientation]
    expected = matrix_to_ints(EXIF_MATRICES[orientation])

    buf = io.BytesIO()
    container = av.open(buf, "w", format="mp4")
    stream = container.add_stream("mpeg4", rate=24)
    stream.width = WIDTH
    stream.height = HEIGHT
    stream.pix_fmt = "yuv420p"
    stream.set_display_rotation(degrees, hflip=hflip, vflip=vflip)
    for i in range(DURATION):
        frame = av.VideoFrame.from_ndarray(
            np.zeros((HEIGHT, WIDTH, 3), dtype=np.uint8), format="rgb24"
        )
        for packet in stream.encode(frame):
            container.mux(packet)
    for packet in stream.encode():
        container.mux(packet)
    container.close()

    buf.seek(0)
    got = _read_matrix(_read_frame(buf))
    if expected == matrix_to_ints(np.eye(3)):
        assert got is None  # identity emits no side data
    else:
        assert got == expected, f"exif {orientation}: wrote {expected}, read {got}"


def test_matrix_and_rotation_setters_are_mutually_exclusive() -> None:
    # Setting one path must clear the other so they don't both apply.
    buf = io.BytesIO()
    with av.open(buf, "w", format="mp4") as container:
        stream = container.add_stream("mpeg4", rate=24)
        stream.width = WIDTH
        stream.height = HEIGHT
        stream.pix_fmt = "yuv420p"
        stream.set_display_rotation(90)
        stream.set_display_matrix(None)  # clears both paths
        frame = av.VideoFrame.from_ndarray(
            np.zeros((HEIGHT, WIDTH, 3), dtype=np.uint8), format="rgb24"
        )
        for packet in stream.encode(frame):
            container.mux(packet)
        for packet in stream.encode():
            container.mux(packet)

    buf.seek(0)
    assert _read_matrix(_read_frame(buf)) is None


def test_set_display_matrix_validates_length() -> None:
    buf = io.BytesIO()
    with av.open(buf, "w", format="mp4") as container:
        stream = container.add_stream("mpeg4", rate=24)
        with pytest.raises(ValueError):
            stream.set_display_matrix([0, 1, 2])


def test_set_display_matrix_none_clears() -> None:
    buf = io.BytesIO()
    with av.open(buf, "w", format="mp4") as container:
        stream = container.add_stream("mpeg4", rate=24)
        stream.set_display_matrix(matrix_to_ints(EXIF_MATRICES[6]))
        stream.set_display_matrix(None)  # clear before encoding
        stream.width = WIDTH
        stream.height = HEIGHT
        stream.pix_fmt = "yuv420p"
        frame = av.VideoFrame.from_ndarray(
            np.zeros((HEIGHT, WIDTH, 3), dtype=np.uint8), format="rgb24"
        )
        for packet in stream.encode(frame):
            container.mux(packet)
        for packet in stream.encode():
            container.mux(packet)

    buf.seek(0)
    assert _read_matrix(_read_frame(buf)) is None



from fractions import Fraction

import av

from .common import fate_suite


def test_chapters() -> None:
    expected = [
        {
            "id": 1,
            "start": 0,
            "end": 5000,
            "time_base": Fraction(1, 1000),
            "metadata": {"title": "start"},
        },
        {
            "id": 2,
            "start": 5000,
            "end": 10500,
            "time_base": Fraction(1, 1000),
            "metadata": {"title": "Five Seconds"},
        },
        {
            "id": 3,
            "start": 10500,
            "end": 15000,
            "time_base": Fraction(1, 1000),
            "metadata": {"title": "Ten point 5 seconds"},
        },
        {
            "id": 4,
            "start": 15000,
            "end": 19849,
            "time_base": Fraction(1, 1000),
            "metadata": {"title": "15 sec - over soon"},
        },
    ]
    path = fate_suite("vorbis/vorbis_chapter_extension_demo.ogg")
    with av.open(path) as container:
        assert container.chapters() == expected


def test_set_chapters() -> None:
    chapters: list[av.container.Chapter] = [
        {
            "id": 1,
            "start": 0,
            "end": 5000,
            "time_base": Fraction(1, 1000),
            "metadata": {"title": "start"},
        }
    ]

    path = fate_suite("h264/interlaced_crop.mp4")
    with av.open(path) as container:
        container.set_chapters(chapters)
        assert container.chapters() == chapters


def test_set_chapters_empty() -> None:
    path = fate_suite("vorbis/vorbis_chapter_extension_demo.ogg")
    with av.open(path) as container:
        assert len(container.chapters()) == 4
        container.set_chapters([])
        assert container.chapters() == []



import io
from fractions import Fraction

import numpy as np
import pytest

import av
import av.datasets

from .common import fate_suite, sandboxed


def test_video_remux() -> None:
    input_path = av.datasets.curated("pexels/time-lapse-video-of-night-sky-857195.mp4")
    output_path = sandboxed("remuxed.mkv")
    input_ = av.open(input_path)
    output = av.open(output_path, "w")

    in_stream = input_.streams.video[0]
    out_stream = output.add_stream_from_template(in_stream)

    for packet in input_.demux(in_stream):
        if packet.size == 0:  # skip the flushing packet, not keyframes with no DTS
            continue

        packet.stream = out_stream
        output.mux(packet)

    input_.close()
    output.close()

    with av.open(output_path) as container:
        # Assert output is a valid media file
        assert len(container.streams.video) == 1
        assert len(container.streams.audio) == 0
        assert container.streams.video[0].codec.name == "h264"

        packet_count = 0
        for packet in container.demux(video=0):
            packet_count += 1

        assert packet_count > 50


def test_add_mux_stream_video() -> None:
    """add_mux_stream creates a video stream without a CodecContext."""
    input_path = av.datasets.curated("pexels/time-lapse-video-of-night-sky-857195.mp4")

    buf = io.BytesIO()
    with av.open(input_path) as input_:
        in_stream = input_.streams.video[0]
        width = in_stream.codec_context.width
        height = in_stream.codec_context.height

        with av.open(buf, "w", format="mp4") as output:
            out_stream = output.add_mux_stream(
                in_stream.codec_context.name, width=width, height=height
            )
            assert out_stream.codec_context is None
            assert out_stream.type == "video"

            out_stream.time_base = in_stream.time_base

            for packet in input_.demux(in_stream):
                if packet.size == 0:
                    continue
                packet.stream = out_stream
                output.mux(packet)

    buf.seek(0)
    with av.open(buf) as result:
        assert len(result.streams.video) == 1
        assert result.streams.video[0].codec_context.width == width
        assert result.streams.video[0].codec_context.height == height


def test_add_mux_stream_no_codec_context() -> None:
    """add_mux_stream streams have no codec context and repr does not crash."""
    buf = io.BytesIO()
    with av.open(buf, "w", format="mp4") as output:
        video_stream = output.add_mux_stream("h264", width=1920, height=1080)
        audio_stream = output.add_mux_stream("aac", rate=44100)

        assert video_stream.codec_context is None
        assert audio_stream.codec_context is None
        # repr should not crash
        assert "video/<nocodec>" in repr(video_stream)
        assert "audio/<nocodec>" in repr(audio_stream)


def test_add_mux_stream_matroska_extradata() -> None:
    """Regression test for #2198.

    Muxing pre-encoded H.264 packets into Matroska used to fail in
    ``avformat_write_header`` because the muxer needs ``CodecPrivate`` (the codec
    extradata) before the first packet is written. When the packets are annex-b
    with in-band parameter sets but the stream has no extradata,
    ``add_mux_stream`` extracts it from the bitstream so the file is written and
    stays decodable.
    """
    if av.codec.Codec("h264", "w").name != "libx264":
        pytest.skip("requires libx264")

    # Encode without a global header, so the packets carry annex-b parameter
    # sets in-band and the codec context exposes no extradata to copy.
    cc = av.CodecContext.create("libx264", "w")
    cc.width, cc.height, cc.pix_fmt = 320, 240, "yuv420p"
    cc.time_base = Fraction(1, 24)
    cc.framerate = Fraction(24, 1)

    packets = []
    for i in range(24):
        frame = av.VideoFrame.from_ndarray(
            np.zeros((240, 320, 3), dtype="uint8"), format="rgb24"
        )
        frame.pts = i
        frame.time_base = Fraction(1, 24)
        packets.extend(cc.encode(frame))
    packets.extend(cc.encode(None))
    assert cc.extradata is None  # nothing to copy onto the stream

    buf = io.BytesIO()
    with av.open(buf, "w", format="matroska") as output:
        out_stream = output.add_mux_stream("libx264", rate=24, width=320, height=240)
        for packet in packets:
            packet.stream = out_stream
            output.mux(packet)

    assert _decoded_frame_count(buf) == 24


def test_add_stream_from_template_copies_time_base() -> None:
    """add_stream_from_template must propagate the source stream's time_base.

    AVCodecParameters does not carry time_base, so without an explicit copy
    the output stream's time_base stays as None
    """
    video_path = av.datasets.curated("pexels/time-lapse-video-of-night-sky-857195.mp4")
    with (
        av.open(video_path) as input_,
        av.open(io.BytesIO(), "w", format="mp4") as output,
    ):
        in_video = input_.streams.video[0]
        out_video = output.add_stream_from_template(in_video)
        assert out_video.time_base is not None
        assert out_video.time_base == in_video.time_base

    audio_path = fate_suite("audio-reference/chorusnoise_2ch_44kHz_s16.wav")
    with (
        av.open(audio_path) as input_,
        av.open(io.BytesIO(), "w", format="wav") as output,
    ):
        in_audio = input_.streams.audio[0]
        out_audio = output.add_stream_from_template(in_audio)
        assert out_audio.time_base is not None
        assert out_audio.time_base == in_audio.time_base


def _make_b_frame_mkv(n: int = 48) -> io.BytesIO:
    """Encode `n` frames with B-frames into an in-memory MKV.

    Matroska stores only presentation timestamps, so when this is demuxed again
    libavformat cannot reconstruct a DTS for the leading reordered packets and
    leaves it as None -- including on the very first packet, which is the
    keyframe. That is exactly the layout that broke the remux example in #1917.
    """
    buf = io.BytesIO()
    with av.open(buf, "w", format="matroska") as out:
        stream = out.add_stream("h264", rate=30)
        stream.width, stream.height, stream.pix_fmt = 160, 120, "yuv420p"
        stream.options = {"bf": "3", "g": "30"}
        for i in range(n):
            img = np.full((120, 160, 3), (i * 5) % 256, dtype="uint8")
            frame = av.VideoFrame.from_ndarray(img, format="rgb24")
            for packet in stream.encode(frame):
                out.mux(packet)
        for packet in stream.encode(None):
            out.mux(packet)
    buf.seek(0)
    return buf


def _decoded_frame_count(buf: io.BytesIO) -> int:
    buf.seek(0)
    with av.open(buf, "r") as container:
        return sum(1 for _ in container.decode(video=0))


def test_remux_keeps_keyframe_with_none_dts() -> None:
    """Regression test for #1917.

    A keyframe can legitimately demux with ``dts is None`` (B-frame stream in a
    PTS-only container like MKV). The remux loop must skip only the empty
    flushing packet (``size == 0``), not every ``dts is None`` packet, otherwise
    the keyframe is dropped and the output is undecodable.
    """
    if av.codec.Codec("h264", "w").name != "libx264":
        pytest.skip("requires libx264")

    source = _make_b_frame_mkv()
    expected_frames = _decoded_frame_count(source)
    assert expected_frames > 0

    # Precondition: the first packet really is a keyframe without a DTS, which is
    # what the old `dts is None` filter would have wrongly discarded.
    source.seek(0)
    with av.open(source, "r") as input_:
        first = next(p for p in input_.demux(input_.streams.video[0]) if p.size)
        assert first.is_keyframe
        assert first.dts is None

    source.seek(0)
    output = io.BytesIO()
    with (
        av.open(source, "r") as input_,
        av.open(output, "w", format="matroska") as out,
    ):
        in_video = input_.streams.video[0]
        out_video = out.add_stream_from_template(in_video)
        for packet in input_.demux(in_video):
            if packet.size == 0:  # the flushing packet, not a keyframe with no DTS
                continue
            packet.stream = out_video
            out.mux(packet)

    # The keyframe survived: every frame still decodes and the first packet of
    # the remuxed stream is a keyframe.
    assert _decoded_frame_count(output) == expected_frames
    output.seek(0)
    with av.open(output, "r") as container:
        first_out = next(p for p in container.demux(video=0) if p.size)
        assert first_out.is_keyframe



import os
import time

import av
import av.datasets
from av.codec.hwaccel import HWAccel, hwdevices_available

# What accelerator to use.
# Recommendations:
#   Windows:
#       - d3d11va (Direct3D 11)
#           * available with built-in ffmpeg in PyAV binary wheels, and gives access to
#             all decoders, but performance may not be as good as vendor native interfaces.
#       - cuda (NVIDIA NVDEC), qsv (Intel QuickSync)
#           * may be faster than d3d11va, but requires custom ffmpeg built with those libraries.
#   Linux (all options require custom FFmpeg):
#       - vaapi (Intel, AMD)
#       - cuda (NVIDIA)
#   Mac:
#       - videotoolbox
#           * available with built-in ffmpeg in PyAV binary wheels, and gives access to
#             all accelerators available on Macs. This is the only option on MacOS.

HW_DEVICE = os.environ["HW_DEVICE"] if "HW_DEVICE" in os.environ else None

if "TEST_FILE_PATH" in os.environ:
    test_file_path = os.environ["TEST_FILE_PATH"]
else:
    test_file_path = av.datasets.curated(
        "pexels/time-lapse-video-of-night-sky-857195.mp4"
    )

if HW_DEVICE is None:
    print(f"Please set HW_DEVICE. Options are: {hwdevices_available()}")
    exit()

assert HW_DEVICE in hwdevices_available(), f"{HW_DEVICE} not available."

print("Decoding in software (auto threading)...")

container = av.open(test_file_path)

container.streams.video[0].thread_type = "AUTO"

start_time = time.time()
frame_count = 0
for packet in container.demux(video=0):
    for _ in packet.decode():
        frame_count += 1

sw_time = time.time() - start_time
sw_fps = frame_count / sw_time
assert frame_count == container.streams.video[0].frames
container.close()

print(
    f"Decoded with software in {sw_time:.2f}s ({sw_fps:.2f} fps).\n"
    f"Decoding with {HW_DEVICE}"
)

hwaccel = HWAccel(device_type=HW_DEVICE, allow_software_fallback=False)

# Note the additional argument here.
container = av.open(test_file_path, hwaccel=hwaccel)

start_time = time.time()
frame_count = 0
for packet in container.demux(video=0):
    for _ in packet.decode():
        frame_count += 1

hw_time = time.time() - start_time
hw_fps = frame_count / hw_time
assert frame_count == container.streams.video[0].frames
container.close()

print(f"Decoded with {HW_DEVICE} in {hw_time:.2f}s ({hw_fps:.2f} fps).")



from av import AudioLayout


def _test_stereo(layout: AudioLayout) -> None:
    assert layout.name == "stereo"
    assert layout.nb_channels == 2
    assert repr(layout) == "<av.AudioLayout 'stereo'>"

    # Re-enable when FFmpeg 6.0 is dropped.

    # assert layout.channels[0].name == "FL"
    # assert layout.channels[0].description == "front left"
    # assert repr(layout.channels[0]) == "<av.AudioChannel 'FL' (front left)>"
    # assert layout.channels[1].name == "FR"
    # assert layout.channels[1].description == "front right"
    # assert repr(layout.channels[1]) == "<av.AudioChannel 'FR' (front right)>"


def test_stereo_from_str() -> None:
    layout = AudioLayout("stereo")
    _test_stereo(layout)


def test_stereo_from_layout() -> None:
    layout2 = AudioLayout(AudioLayout("stereo"))
    _test_stereo(layout2)



import argparse
import logging
import json
import os
import platform
import subprocess
import time


def get_platform():
    # Allow forcing the target platform so we can fetch e.g. an armv7l build
    # while running on an x86_64 host (cross-compilation).
    forced = os.environ.get("PYAV_VENDOR_PLATFORM")
    if forced:
        return forced

    system = platform.system()
    machine = platform.machine().lower()
    is_arm64 = machine in {"arm64", "aarch64"}
    if system == "Linux":
        prefix = "manylinux-" if platform.libc_ver()[0] == "glibc" else "musllinux-"
        return prefix + machine
    elif system == "Darwin":
        return "macos-arm64" if is_arm64 else "macos-x86_64"
    elif system == "Windows":
        return "windows-aarch64" if is_arm64 else "windows-x86_64"
    else:
        return "unknown"


parser = argparse.ArgumentParser(description="Fetch and extract tarballs")
parser.add_argument("destination_dir")
parser.add_argument("--cache-dir", default="tarballs")
parser.add_argument("--config-file", default=os.path.splitext(__file__)[0] + ".json")
args = parser.parse_args()
logging.basicConfig(level=logging.INFO)

with open(args.config_file) as fp:
    config = json.load(fp)

# ensure destination directory exists
logging.info(f"Creating directory {args.destination_dir}")
if not os.path.exists(args.destination_dir):
    os.makedirs(args.destination_dir)

tarball_url = config["url"].replace("{platform}", get_platform())

# download tarball
tarball_name = tarball_url.split("/")[-1]
tarball_file = os.path.join(args.cache_dir, tarball_name)
if not os.path.exists(tarball_file):
    logging.info(f"Downloading {tarball_url}")
    if not os.path.exists(args.cache_dir):
        os.mkdir(args.cache_dir)
    subprocess.check_call(
        ["curl", "--location", "--output", tarball_file, "--silent", tarball_url]
    )

logging.info(f"Extracting {tarball_name}")
subprocess.check_call(["tar", "-C", args.destination_dir, "-xf", tarball_file])

# Some tarball members carry pre-1980 mtimes, which the ZIP format (and thus
# delvewheel's wheel repackaging) cannot represent. Bump any such file to now.
ZIP_EPOCH = 315532800  # 1980-01-01 00:00:00 UTC
now = time.time()
for root, _, files in os.walk(args.destination_dir):
    for name in files:
        path = os.path.join(root, name)
        if os.path.getmtime(path) < ZIP_EPOCH:
            os.utime(path, (now, now))



from fractions import Fraction

import av

from .common import TestCase, fate_suite


class TestAudioProbe(TestCase):
    def setUp(self):
        self.file = av.open(fate_suite("aac/latm_stereo_to_51.ts"))

    def test_container_probing(self) -> None:
        assert self.file.bit_rate == 269558
        assert self.file.duration == 6165333
        assert str(self.file.format) == "<av.ContainerFormat 'mpegts'>"
        assert self.file.format.name == "mpegts"
        assert self.file.format.long_name == "MPEG-TS (MPEG-2 Transport Stream)"
        assert self.file.metadata == {}
        assert self.file.size == 207740
        assert self.file.start_time == 1400000
        assert len(self.file.streams) == 1

    def test_stream_probing(self) -> None:
        stream = self.file.streams[0]

        assert isinstance(stream, av.AudioStream)
        assert str(stream).startswith(
            "<av.AudioStream #0 aac_latm at 48000Hz, stereo, fltp at "
        )

        # actual stream properties
        assert stream.duration == 554880
        assert stream.frames == 0
        assert stream.id == 256
        assert stream.index == 0
        assert stream.language == "eng"
        assert stream.metadata == {"language": "eng"}
        assert stream.profile == "LC"
        assert stream.start_time == 126000
        assert stream.time_base == Fraction(1, 90000)
        assert stream.type == "audio"

        # codec context properties
        assert stream.bit_rate is None
        assert stream.channels == 2
        assert stream.codec.name == "aac_latm"
        assert stream.codec.long_name == "AAC LATM (Advanced Audio Coding LATM syntax)"
        assert stream.format.bits == 32
        assert stream.format.name == "fltp"
        assert stream.layout.name == "stereo"
        assert stream.max_bit_rate is None
        assert stream.sample_rate == 48000


class TestAudioProbeCorrupt(TestCase):
    def setUp(self) -> None:
        # write an empty file
        path = self.sandboxed("empty.flac")
        with open(path, "wb"):
            pass

        self.file = av.open(path, "r")

    def test_container_probing(self) -> None:
        assert self.file.bit_rate == 0
        assert self.file.duration is None
        assert str(self.file.format) == "<av.ContainerFormat 'flac'>"
        assert self.file.format.name == "flac"
        assert self.file.format.long_name == "raw FLAC"
        assert self.file.metadata == {}
        assert self.file.size == 0
        assert self.file.start_time is None
        assert len(self.file.streams) == 1

    def test_stream_probing(self) -> None:
        stream = self.file.streams[0]

        assert isinstance(stream, av.AudioStream)
        assert str(stream).startswith(
            "<av.AudioStream #0 flac at 0Hz, 0 channels, None at "
        )

        # actual stream properties
        assert stream.duration is None
        assert stream.frames == 0
        assert stream.id == 0
        assert stream.index == 0
        assert stream.language is None
        assert stream.metadata == {}
        assert stream.profile is None
        assert stream.start_time is None
        assert stream.time_base == Fraction(1, 90000)
        assert stream.type == "audio"

        # codec context properties
        assert stream.bit_rate is None
        assert stream.codec.name == "flac"
        assert stream.codec.long_name == "FLAC (Free Lossless Audio Codec)"
        assert stream.channels == 0
        assert stream.format is None
        assert stream.layout.name == "0 channels"
        assert stream.max_bit_rate is None
        assert stream.sample_rate == 0


class TestDataProbe(TestCase):
    def setUp(self) -> None:
        self.file = av.open(fate_suite("mxf/track_01_v02.mxf"))

    def test_container_probing(self) -> None:
        assert self.file.bit_rate == 27872687
        assert self.file.duration == 417083
        assert str(self.file.format) == "<av.ContainerFormat 'mxf'>"
        assert self.file.format.name == "mxf"
        assert self.file.format.long_name == "MXF (Material eXchange Format)"
        assert self.file.size == 1453153
        assert self.file.start_time == 0
        assert len(self.file.streams) == 4

        for key, value in (
            ("application_platform", "AAFSDK (MacOS X)"),
            ("comment_Comments", "example comment"),
            ("comment_UNC Path", "/Users/mark/Desktop/dnxhr_tracknames_export.aaf"),
            ("company_name", "Avid Technology, Inc."),
            ("generation_uid", "b6bcfcab-70ff-7331-c592-233869de11d2"),
            ("material_package_name", "Example.new.04"),
            (
                "material_package_umid",
                "0x060A2B340101010101010F001300000057E19D16BA8202DB060E2B347F7F2A80",
            ),
            ("modification_date", "2016-09-20T20:33:26.000000Z"),
            ("operational_pattern_ul", "060e2b34.04010102.0d010201.10030000"),
            ("product_name", "Avid Media Composer 8.6.3.43955"),
            ("product_uid", "acfbf03a-4f42-a231-d0b7-c06ecd3d4ad7"),
            ("product_version", "Unknown version"),
            ("project_name", "UHD"),
            ("uid", "4482d537-4203-ea40-9e4e-08a22900dd39"),
        ):
            assert self.file.metadata.get(key) == value

    def test_stream_probing(self) -> None:
        stream = self.file.streams[0]

        assert str(stream).startswith("<av.DataStream #0 data/<nocodec> at ")

        assert stream.duration == 37537
        assert stream.frames == 0
        assert stream.id == 1
        assert stream.index == 0
        assert stream.language is None
        assert stream.metadata == {
            "data_type": "video",
            "file_package_umid": "0x060A2B340101010101010F001300000057E19D16BA8302DB060E2B347F7F2A80",
            "track_name": "Base",
        }
        assert stream.profile is None
        assert stream.start_time == 0
        assert stream.time_base == Fraction(1, 90000)
        assert stream.type == "data"
        assert not hasattr(stream, "codec")


class TestSubtitleProbe(TestCase):
    def setUp(self) -> None:
        self.file = av.open(fate_suite("sub/MovText_capability_tester.mp4"))

    def test_container_probing(self) -> None:
        assert self.file.bit_rate == 810
        assert self.file.duration == 8140000
        assert str(self.file.format) == "<av.ContainerFormat 'mov,mp4,m4a,3gp,3g2,mj2'>"
        assert self.file.format.name == "mov,mp4,m4a,3gp,3g2,mj2"
        assert self.file.format.long_name == "QuickTime / MOV"
        assert self.file.metadata == {
            "compatible_brands": "isom",
            "creation_time": "2012-07-04T05:10:41.000000Z",
            "major_brand": "isom",
            "minor_version": "1",
        }
        assert self.file.size == 825
        assert self.file.start_time is None
        assert len(self.file.streams) == 1

    def test_stream_probing(self) -> None:
        stream = self.file.streams[0]

        assert str(stream).startswith("<av.SubtitleStream #0 subtitle/mov_text at ")

        # actual stream properties
        assert stream.duration == 8140
        assert stream.frames == 6
        assert stream.id == 1
        assert stream.index == 0
        assert stream.language == "und"
        assert stream.metadata == {
            "creation_time": "2012-07-04T05:10:41.000000Z",
            "handler_name": "reference.srt - Imported with GPAC 0.4.6-DEV-rev4019",
            "language": "und",
        }
        assert stream.profile is None
        assert stream.start_time is None
        assert stream.time_base == Fraction(1, 1000)
        assert stream.type == "subtitle"

        # codec context properties
        assert stream.codec.name == "mov_text"
        assert stream.codec.long_name == "3GPP Timed Text subtitle"


class TestVideoProbe(TestCase):
    def setUp(self) -> None:
        self.file = av.open(fate_suite("mpeg2/mpeg2_field_encoding.ts"))

    def test_container_probing(self) -> None:
        assert self.file.bit_rate == 3950617
        assert self.file.duration == 1620000
        assert str(self.file.format) == "<av.ContainerFormat 'mpegts'>"
        assert self.file.format.name == "mpegts"
        assert self.file.format.long_name == "MPEG-TS (MPEG-2 Transport Stream)"
        assert self.file.metadata == {}
        assert self.file.size == 800000
        assert self.file.start_time == 22953408322
        assert len(self.file.streams) == 1

    def test_stream_probing(self) -> None:
        stream = self.file.streams[0]

        assert isinstance(stream, av.video.stream.VideoStream)
        assert str(stream).startswith(
            "<av.VideoStream #0 mpeg2video, yuv420p 720x576 at "
        )

        # actual stream properties
        assert stream.average_rate == Fraction(25, 1)
        assert stream.duration == 145800
        assert stream.frames == 0
        assert stream.id == 4131
        assert stream.index == 0
        assert stream.language is None
        assert stream.metadata == {}
        assert stream.profile == "Simple"
        assert stream.start_time == 2065806749
        assert stream.time_base == Fraction(1, 90000)
        assert stream.type == "video"

        # codec context properties
        assert stream.bit_rate == 3364800
        assert stream.codec.long_name == "MPEG-2 video"
        assert stream.codec.name == "mpeg2video"
        assert stream.display_aspect_ratio == Fraction(4, 3)
        assert stream.format.name == "yuv420p"
        assert not stream.has_b_frames
        assert stream.height == 576
        assert stream.max_bit_rate is None
        assert stream.sample_aspect_ratio == Fraction(16, 15)
        assert stream.width == 720
        assert stream.coded_width == 0
        assert stream.coded_height == 0

        assert not hasattr(stream, "framerate")
        assert not hasattr(stream, "rate")


class TestVideoProbeCorrupt(TestCase):
    def setUp(self) -> None:
        path = self.sandboxed("empty.h264")
        with open(path, "wb"):
            pass

        self.file = av.open(path)

    def test_container_probing(self) -> None:
        assert str(self.file.format) == "<av.ContainerFormat 'h264'>"
        assert self.file.format.name == "h264"
        assert self.file.format.long_name == "raw H.264 video"
        assert self.file.size == 0
        assert self.file.bit_rate == 0
        assert self.file.duration is None

        assert len(self.file.streams) == 1
        assert self.file.start_time is None
        assert self.file.metadata == {}

    def test_stream_probing(self) -> None:
        stream = self.file.streams[0]

        assert isinstance(stream, av.VideoStream)
        assert str(stream).startswith("<av.VideoStream #0 h264, None 0x0 at ")

        # actual stream properties
        assert stream.type == "video"
        assert stream.duration is None
        assert stream.frames == 0
        assert stream.id == 0
        assert stream.index == 0
        assert stream.language is None
        assert stream.metadata == {}
        assert stream.profile is None
        assert stream.start_time is None
        assert stream.time_base == Fraction(1, 1200000)

        # codec context properties
        assert stream.bit_rate is None
        assert stream.codec.long_name == "H.264 / AVC / MPEG-4 AVC / MPEG-4 part 10"
        assert stream.codec.name == "h264"
        assert stream.display_aspect_ratio is None
        assert stream.format is None
        assert not stream.has_b_frames
        assert stream.height == 0
        assert stream.max_bit_rate is None
        assert stream.sample_aspect_ratio is None
        assert stream.width == 0

        assert stream.coded_width == 0
        assert stream.coded_height == 0
